help = {
    'select':'pairtools select CONDITION : A Python expression; if it returns True, select the read pair. \
             Any column declared in the #columns line of the pairs header can be accessed \
            by its name. If the header lacks the #columns line, the columns are assumed \
            to follow the .pairs/.pairsam standard (readID, chrom1, chrom2, pos1, pos2, \
            strand1, strand2, pair_type). Finally, CONDITION has access to COLS list\
            which contains the string values of columns. In Bash, quote CONDITION with\
            single quotes, and use double quotes for string variables inside CONDITION.',
    'add_columns':'Report extra columns describing alignments \
                    Possible values (can take multiple values as \
                    a comma-separated list): a SAM tag (any pair \
                    of uppercase letters) or mapq, pos5, pos3, \
                    cigar, read_len, matched_bp, algn_ref_span, \
                    algn_read_span, dist_to_5, dist_to_3, seq, \
                    mismatches',
    'max_mismatch':'pairtools dedup:Pairs with both sides mapped within this\
                distance (bp) from each other are considered\
                    duplicates. [dedup option]  [default: 3'
}


# -*- coding: utf-8 -*-
import umap
import numpy as np
from sklearn.manifold import MDS, TSNE, LocallyLinearEmbedding, Isomap
from scipy.stats import norm
def gaussian_distance(self, matrix1, matrix2, sigma):
        # 计算高斯核矩阵
        kernel1 = np.exp(-np.sum((matrix1[:, np.newaxis] - matrix1) ** 2, axis=2) / (2 * sigma ** 2))
        kernel2 = np.exp(-np.sum((matrix2[:, np.newaxis] - matrix2) ** 2, axis=2) / (2 * sigma ** 2))

        # 计算高斯距离
        distance = np.sum((kernel1 - kernel2) ** 2)

        return distance


def cosine_similarity(self, matrix_A, matrix_B):
        ####计算两个矩阵的余弦相似性
        vector_A = matrix_A.ravel()  # 使用ravel()代替flatten()，无需创建副本
        vector_B = matrix_B.ravel()

        dot_product = np.dot(vector_A, vector_B)
        norm_A = np.linalg.norm(vector_A)
        norm_B = np.linalg.norm(vector_B)

        similarity = dot_product / (norm_A * norm_B)
        return similarity
# PCA :---------------------------------------------------
def scale_matrix(self, matrix, target_size):
        n = matrix.shape[0]  # 原始矩阵的大小
        scale_factor = n / target_size  # 计算缩放比例

        # 构建目标矩阵的行和列索引
        row_indices = np.arange(target_size) * scale_factor
        col_indices = np.arange(target_size) * scale_factor

        # 使用索引进行插值计算
        row_floor = row_indices.astype(int)
        row_ceil = np.ceil(row_indices).astype(int)
        col_floor = col_indices.astype(int)
        col_ceil = np.ceil(col_indices).astype(int)

        # 计算四个邻近像素值
        Q11 = matrix[row_floor[:, None], col_floor]
        Q12 = matrix[row_floor[:, None], col_ceil]
        Q21 = matrix[row_ceil[:, None], col_floor]
        Q22 = matrix[row_ceil[:, None], col_ceil]

        # 双线性插值计算目标矩阵
        target_matrix = (Q11 * (row_ceil[:, None] - row_indices[:, None]) * (col_ceil - col_indices) +
                         Q21 * (row_indices[:, None] - row_floor[:, None]) * (col_ceil - col_indices) +
                         Q12 * (row_ceil[:, None] - row_indices[:, None]) * (col_indices - col_floor) +
                         Q22 * (row_indices[:, None] - row_floor[:, None]) * (col_indices - col_floor))

        return target_matrix
        
def PCAEmbedding(X, dim=2):
    """
    
    Principal components analysis，PCA.
    
    
    Parameters
    ----------
    X : numpy.ndarray
        Coordinates of input data points.
        
    dim : int, optional
        The dimension of the projected points. 
        The default is 2.


    Returns
    -------
    Y : numpy.ndarray
        Coordinates of the projected points.

    """
    
    X=X-np.mean(X,axis=0)
    U, S, V = np.linalg.svd(X, full_matrices=False, compute_uv=True)
    Y = np.dot(V[:dim,:],X.T).T
    return(Y)
    

def MDSEmbedding(X, dim=2):
    """
    Multidimensional Scaling, MDS.

    Parameters
    ----------
    X : numpy.ndarray
        Coordinates of input data points.

    dim : int, optional
        The dimension of the projected points.
        The default is 2.

    Returns
    -------
    Y : numpy.ndarray
        Coordinates of the projected points.

    """
    mds = MDS(n_components=dim, normalized_stress='auto')
    Y = mds.fit_transform(X)

    return Y


def tSNEEmbedding(X, dim=2, perplexity=30.0, learning_rate=200.0, n_iter=1000):
    """
    t-Distributed Stochastic Neighbor Embedding，t-SNE.

    Parameters
    ----------
    X : numpy.ndarray
        Coordinates of input data points.

    dim : int, optional
        The dimension of the projected points.
        The default is 2.

    perplexity : float, optional
        The perplexity parameter for t-SNE.
        The default is 30.0.

    learning_rate : float, optional
        The learning rate for t-SNE optimization.
        The default is 200.0.

    n_iter : int, optional
        The number of iterations for t-SNE optimization.
        The default is 1000.

    Returns
    -------
    Y : numpy.ndarray
        Coordinates of the projected points.
    """
    tsne = TSNE(n_components=dim, perplexity=perplexity, learning_rate=learning_rate, n_iter=n_iter, method='exact')
    Y = tsne.fit_transform(X)
    return Y


def UMAPEmbedding(X, dim=2):
    """
    Uniform Manifold Approximation and Projection (UMAP).

    Parameters
    ----------
    X : numpy.ndarray
        Coordinates of input data points.

    dim : int, optional
        The dimension of the projected points.
        The default is 2.

    Returns
    -------
    Y : numpy.ndarray
        Coordinates of the projected points.
    """

    reducer = umap.UMAP(n_components=dim)
    Y = reducer.fit_transform(X)

    return Y


def LLEEmbedding(X, dim=2):
    """
    Locally Linear Embedding (LLE).

    Parameters
    ----------
    X : numpy.ndarray
        Coordinates of input data points.

    dim : int, optional
        The dimension of the projected points.
        The default is 2.

    Returns
    -------
    Y : numpy.ndarray
        Coordinates of the projected points.
    """

    lle = LocallyLinearEmbedding(n_components=dim)
    Y = lle.fit_transform(X)

    return Y


def IsomapEmbedding(X, dim=2, n_neighbors=5):
    """
    Isometric Mapping (Isomap).

    Parameters
    ----------
    X : numpy.ndarray
        Coordinates of input data points.

    dim : int, optional
        The dimension of the projected points.
        The default is 2.

    n_neighbors : int, optional
        The number of neighbors to consider for each data point.
        The default is 5.

    Returns
    -------
    Y : numpy.ndarray
        Coordinates of the projected points.
    """

    isomap = Isomap(n_components=dim, n_neighbors=n_neighbors)
    Y = isomap.fit_transform(X)

    return Y













import pandas as pd

def generate_barcode(index, pcr_file, tn5_file,out):
    # 定义序列
    seq1 = "AATGATACGGCGACCACCGAGATCT"
    seq2 = "TCGTCGGCAGCGTC"
    seq3 = "AGATGTGTATAAGAGACAG"
    
    # 读取PCR和TN5的条形码文件
    PCR = pd.read_csv(pcr_file, sep='\t', header=None, index_col=0)
    TN5 = pd.read_csv(tn5_file, sep='\t', header=None, index_col=0)
    
    with open(out, 'w') as output:
        for line in index:
            a = line.strip().split()
            if a[0].isdigit():
                PCR_BC = int(a[0])
                BC_start = int(a[1]) if len(a) > 1 else 1
                BC_end = int(a[2]) if len(a) > 2 else 24
                PCR_Name = f'P{PCR_BC}'

                PCR_Seq = PCR.loc[PCR_BC, 1]
                for TN5_BC in range(BC_start, BC_end + 1):
                    TN5_Seq = TN5.loc[TN5_BC, 1]
                    Name = f'{PCR_Name}B{TN5_BC}'
                    BARCODE = f'{seq1}{PCR_Seq}{seq2}{TN5_Seq}{seq3}'
                    
                    # print(f"Name: {Name}")
                    # print(f"PCR_Seq: {PCR_Seq}")
                    # print(f"TN5_Seq: {TN5_Seq}")
                    # print("==============================================================")

                    output.write(f">{Name}\n")
                    output.write(f"{BARCODE}\n")
            else:
                print("Barcode is not an integer!")
                break


import os
import sys
import subprocess
import time
import re
from Bio.SeqIO.QualityIO import FastqGeneralIterator
from Bio import SeqIO
from Levenshtein import hamming
import gzip
import logging
# jar_dir = sys.prefix + '/inHic' 
# jar_path = jar_dir + '/utils/BarcodeIdentification_v1.2.0.jar'
jar_path = '/cluster/home/Kangwen/Hic/seq/4_scSPRITE/Code/BarcodeIdentification_v1.2.0.jar'



def run_cmd_return_time(cmd):
    t = time.time()
    logging.debug(cmd)
    subprocess.run(cmd, shell=True, executable="/bin/bash")
    return time.time() - t


def sprite_filter_reads(infastq,outfastq):
    fastq = gzip.open(outfastq,'w+')
    with gzip.open(infastq, "rt") as fp:
        flag = 0
        for line in fp:
            #line = line.decode()
            if '@' in line and '[NOT_FOUND]' not in line and ']' in line[-2]:
                flag = 3
                fastq.write(line.encode('utf-8'))
                continue

            while flag !=0:
                flag-=1
                fastq.write(line.encode('utf-8'))
                break
    fastq.close()

def fastp(fq1,fq2,o1,o2, threads,max=True,adapter:list=None):
    cmd = 'fastp --overrepresentation_analysis'
    if adapter is not None:
        cmd += ' --adapter_sequence ' + adapter[0]
        cmd += ' --adapter_sequence_r2 ' + adapter[1]
    cmd += ' --correction --thread ' + str(threads)
    cmd += ' --html trimmed.fastp.html '
    cmd += ' --json trimmed.fastp.json '
    cmd += ' -i ' + fq1 + ' -I ' + fq2
    cmd += ' --out1 %s --out2 %s' % (o1, o2)
    cmd += ' --failed_out failed.fq.gz'
    if max:
        cmd += ' --max_len1 100'
        cmd += ' --max_len2 100'
    return run_cmd_return_time(cmd)

def samtools_qc(threads,qc,fastq):
    if qc == 0:
         cmd = 'cp %s.bam %s.qc.bam' % (fastq,fastq)
    else:
        cmd = 'samtools view -h -@ %s -b -q %s %s.bam > %s.qc.bam' % (threads, qc, fastq,fastq)
    
    return run_cmd_return_time(cmd)

def bwa(fa, threads,fastq):
    cmd = 'bwa mem ' + fa
    cmd += ' -t ' + str(threads)
    cmd += ' trimmed-pair1.fastq.gz trimmed-pair2.fastq.gz > '  + fastq + '.sam'
    return run_cmd_return_time(cmd)

def bowtie2(fa, threads,fastq):
    cmd = 'bowtie2 -p ' + str(threads)
    cmd += ' -x ' + fa
    cmd += ' -1 trimmed-pair1.fastq.gz -2 trimmed-pair2.fastq.gz -S ' + fastq + '.sam'
    return run_cmd_return_time(cmd)

def sam_to_bam(fastq, threads):
    
    cmd = 'samtools view -@ ' + str(threads) + '-S ' + fastq + '.sam -b -o ' + fastq + '.bam'
    return run_cmd_return_time(cmd)

def pairtools_parse(fastq, genomesize, species, add_columns, threads):

    cmd = 'pairtools parse ' + fastq + '.qc.bam ' + ' -c ' + genomesize
    cmd += ' --drop-sam --drop-seq --output-stats ' + fastq + '.stats '
    cmd += ' --assembly ' + species  + ' --no-flip '
    cmd += ' --add-columns ' + add_columns
    cmd += ' --walks-policy all '
    cmd += '-o ' + fastq + '.pairs.gz'
    cmd += ' --nproc-in ' + str(threads) +  ' --nproc-out ' + str(threads)
    return run_cmd_return_time(cmd)

def pairtools_restrict(srr, enzyme):
    cmd = 'pairtools restrict -f %s %s.pairs.gz -o %s.restrict.pairs.gz' % (enzyme, srr, srr)
    return run_cmd_return_time(cmd)

def pairtools_select(srr, select):
    cmd = 'pairtools select ' + '\"' + select + '\"'
    cmd += ' ' + srr + '.restrict.pairs.gz -o ' + srr+ '.select.pairs.gz'
    return run_cmd_return_time(cmd)

def pairtools_sort(fastq):
    cmd = 'pairtools sort ' + fastq + '.select.pairs.gz -o ' + fastq +'.sort.pairs.gz'
    return run_cmd_return_time(cmd)

def pairtools_dedup(fastq, max_mismatch):
    cmd = 'pairtools dedup '
    cmd += '--max-mismatch ' + max_mismatch
    cmd += ' --mark-dups --output ' + '>( pairtools split --output-pairs ' + fastq + \
        '.dedup.pairs.gz --output-sam ' + fastq + \
        '.nodups.bam ) --output-unmapped >( pairtools split --output-pairs  ' + fastq + \
        '.unmapped.pairs.gz --output-sam  ' + fastq + \
        '.unmapped.bam ) --output-dups >( pairtools split --output-pairs  ' + fastq + \
        '.dups.pairs.gz --output-sam ' + fastq + \
        '.dups.bam ) --output-stats  ' + fastq + \
        '.dedup.stats ' + fastq + \
        '.sort.pairs.gz'

    return run_cmd_return_time(cmd)

def cooler_cload_pairs(genomesize, res, fastq, prefix=''):
    cmd = 'cooler cload pairs -c1 2 -p1 3 -c2 4 -p2 5 '
    if prefix is not  None:
        cmd += genomesize + ':' + str(res) + ' ' + fastq + '.%s.pairs.gz '%prefix+ fastq + '_' + str(res) + '.cool'
    else:
        cmd += genomesize + ':' + str(res) + ' ' + fastq + '.pairs.gz '+ fastq + '_' + str(res) + '.cool'
    return run_cmd_return_time(cmd)

def KR_correctMatrix(fastq, res):
    cmd = 'hicCorrectMatrix correct --matrix  ' + fastq + '_' + str(res) + '.cool '
    cmd += ' --correctionMethod KR --outFileName '
    cmd += ' ' + fastq + '_' + str(res) + '.KR.cool'
    cmd += ' --filterThreshold -1.5 5.0 --chromosomes chr1 chr2 chr3 chr4 chr5 chr6 chr7 chr8 chr9 chr10 chr11 chr12 chr13 chr14 chr15 chr16 chr17 chr18 chr19'
    return run_cmd_return_time(cmd)

def cooelr_zoomify(fastq,res, resz):
    cmd = 'cooler zoomify '  + \
        fastq + '_' +str(res) + \
        '.cool -r ' + resz + ' -o ' + fastq + '.mcool'
    return run_cmd_return_time(cmd)

def find_substring_in_long_string(long_string, substrings):
    long_string = str(long_string)
    pattern = re.compile('|'.join(map(re.escape, substrings)))
    match = pattern.search(long_string)
    if match:
        return match.group()
    else:
        return None

def mv_Result():
    print(os.getcwd())
    for val in os.listdir('./'):
        if val.endswith('.mcool') or val.endswith('.cool'):
            os.system('mv %s Result/' % val)
        if val.endswith('.dedup.pairs.gz') or val.endswith('.restrict.pairs.gz'):
            os.system('mv %s Result/' % val)

def GetFastqIndex(srr ,Fastq1_read, Fastq2_read, Barcode1_read, Barcode2_read):
    t = time.time()
 
    Index_name = 'barcode.index'
    All_reads_count = 0
    Barcoded_Reads_count = 0
    MisBarcode_reads_count = 0
   
    fastq1_barcode = gzip.open(srr + "._1_barcode.fastq.gz", "w")
    fastq2_barcode = gzip.open(srr + "._2_barcode.fastq.gz", "w")
    Index_File = open(Index_name, "w+")

    Fastq1_reads = gzip.open(Fastq1_read, "rt")
    Fastq2_reads = gzip.open(Fastq2_read, "rt") 
    Barcode1_reads =  gzip.open(Barcode1_read, "rt")
    Barcode2_reads = gzip.open(Barcode2_read, "rt") 


    for record1, Barcode1, record2, Barcode2 in zip(SeqIO.parse(Fastq1_reads, "fastq"),
                                                    SeqIO.parse(Barcode1_reads, "fastq"),
                                                    SeqIO.parse(Fastq2_reads, "fastq"),
                                                    SeqIO.parse(Barcode2_reads, "fastq")):
  
        # 在这里，你可以使用 record1、Barcode1、record2、Barcode2 来处理每个文件的记录
        All_reads_count += 1
    
        ####Check Read ID
        if (record1.id != record2.id) or (Barcode1.id != Barcode2.id) or (record1.id != Barcode2.id):
            print("The Fastq Files are not matched")

        ####Check and Select barcode
        if len(str(Barcode1.seq)) != 8 or len(str(Barcode2.seq)) != 8 or "N" in str(Barcode1.seq) or "N" in str(Barcode2.seq):
            MisBarcode_reads_count+=1
            continue
        # print(1)
        Barcoded_Reads_count+=1
        ### Write Read ID
        fastq1_barcode.write(("@"+str(record1.id)+"\t"+str(Barcode1.seq)+"_"+str(Barcode2.seq)+"\n").encode("utf-8"))
        fastq2_barcode.write(("@"+str(record2.id)+"\t"+str(Barcode1.seq)+"_"+str(Barcode2.seq)+"\n").encode("utf-8"))
        Index_File.write((str(record2.id)+"\t"+str(Barcode1.seq)+"_"+str(Barcode2.seq)+"\n"))

        ### Write Seq
        fastq1_barcode.write((str(record1.seq)+"\n").encode("utf-8"))
        fastq2_barcode.write((str(record2.seq)+"\n").encode("utf-8"))

        ### Write append Info
        fastq1_barcode.write(("+" + str(record1.id)+"\n").encode("utf-8"))
        fastq2_barcode.write(("+" + str(record2.id)+"\n").encode("utf-8"))

        ### Write quality
        Read1_Quality = ''.join([chr(phred + 33) for phred in record1.letter_annotations["phred_quality"]])
        Read2_Quality = ''.join([chr(phred + 33) for phred in record2.letter_annotations["phred_quality"]])
        fastq1_barcode.write((str(Read1_Quality)+"\n").encode("utf-8"))
        fastq2_barcode.write((str(Read2_Quality)+"\n").encode("utf-8"))

    fastq1_barcode.close()
    fastq2_barcode.close()
    Index_File.close()
    
    print("All_reads_count: ", All_reads_count)
    print("Barcoded_Reads_count: ", Barcoded_Reads_count)
    print("MisBarcode_reads_count", MisBarcode_reads_count)
    return time.time() - t

def digest(size,fa,  enzyme):
    if ' ' in enzyme:
        enzyme = enzyme.replace(' ', '\ ')
    cmd = 'cooler digest  -o digest.bed %s %s %s' % (size, fa, enzyme)
    print(cmd)
    return run_cmd_return_time(cmd)

def split_cells(fastq):
    print(fastq)
    pair = fastq +   '.restrict.pairs.gz'
    ReadDict, SClist = GenerateSCDict('barcode.index')
    GenerateSCPairs(ReadDict, SClist, pair, './Result/SCpair')

def split_fastqs(r1, r2, r1_o, r2_o, barcodes):
	'''Given paired-end fastq data, split reads based off of an inline 10 (or 11) mer barcode'''
	mismatch = 0
	not_found_R1 = 0
	not_found_R2 = 0
	reads = 0
	#fqr1 = FastqGeneralIterator(r1)
	#fqr2 = FastqGeneralIterator(r2)
	#seqzip = zip(fqr1, fqr2) #Zip up the two iterators for expediency

	with gzip.open(r1, "rt") as handle1:
		with gzip.open(r2, "rt") as handle2:
			for (title1, seq1, qual1), (title2, seq2, qual2) in zip(FastqGeneralIterator(handle1),FastqGeneralIterator(handle2)):
				barcode1 = seq1[:8] #Just look in read 1--barcodes SHOULD be the same on both ends of the molecule)
				barcode2 = seq2[:8] #Check barcode 2 as well; print out how many times they disagree
				test1 = checkHamming(barcodes, barcode1)
				test2 = checkHamming(barcodes, barcode2)
				if test1[0]:
					if test2[0]:
					#if the barcodes match, print out the trimmed / split reads to new files
						if test1[1] == test2[1]:
							print("@%s&%s\n%s\n+\n%s" % (title1,test1[1], seq1[11:], qual1[11:]), file=r1_o)
							print("@%s&%s\n%s\n+\n%s" % (title2,test2[1], seq2[11:], qual2[11:]), file=r2_o)
						else:
							mismatch += 1
					else: #If there isn't a match in R1
						not_found_R2 += 1
				elif test2[0]:
					not_found_R1 += 1
				else:
					not_found_R1 += 1
					not_found_R2 += 1
				reads += 1
	print("mismatch: ",mismatch)
	print("not_found_R1: ",not_found_R1)
	print("not_found_R2: ",not_found_R2)
	print("reads: ",reads)
	handle1.close()
	handle2.close()
	return mismatch, not_found_R1, not_found_R2, reads

def GenerateSCDict(ReadIndex):
    ReadDict = {}
    SCset = set()
    count = 0
    startTime = time.time()
    with open(ReadIndex, "r") as file:
        indexInfo = file.readlines()
        print("index has loaded")
        for line in indexInfo:
            IndexComponent = line.rstrip().split("\t")
            barcode = IndexComponent[1]
            ReadDict[IndexComponent[0]] = barcode
            SCset.add(barcode)
            count+=1
            if count%1000000 == 0:
                print("%d index has processed in %f s"%(count, time.time() - startTime))
                startTime = time.time()

    SClist = list(SCset)
    return ReadDict, SClist

def GenerateSCPairs(ReadDict, SClist, pairFile, ResultDir):
    SCPairs = {barcode: [] for barcode in SClist}
    HeadInfo = []
    count = 0
    print("*****Start make single cell pairs*****")
    with gzip.open(pairFile, "rt") as pf:
        startTime = time.time()
        for line in pf:
            if "#" in line:
                HeadInfo.append(line)
                continue
            PairComponent = line.split("\t")
            barcode = ReadDict[PairComponent[0]]
            SCPairs[barcode].append(line)
            count += 1
            if count % 100000 == 0:
                print("%d pairs has processed in %f s" % (count, time.time() - startTime))
                startTime = time.time()

    print("*****Start generate single cell pairs files*****")
    cell_count = 0
    scPairsSummary = open("singleCell_HIC_pairs_summary.txt", "w")
    scPairsSummary.write("Cell Num" + "\t" + "Cell Barcode" + "\t" + "Pairs Count" + "\n")
    for i in range(len(SClist)):
        if len(SCPairs[SClist[i]])>1:
            scPairsSummary.write(str(i+1)+"\t"+str(SClist[i])+"\t"+str(len(SCPairs[SClist[i]]))+"\n")
        # if len(SCPairs[SClist[i]])<1000:
        #     continue
        cell_count+=1
        print("Process cell %d,\t"%(cell_count),end='')
        if not os.path.exists(ResultDir):
            os.mkdir(ResultDir)
        fp = gzip.open(ResultDir+"/cell_"+str(cell_count)+"_"+str(i)+".pairs.gz","w+")
        for singel_head in HeadInfo:
            fp.write(singel_head.encode('utf-8'))
        count=0

        for pair in SCPairs[SClist[i]]:
            fp.write(pair.encode('utf-8'))
            count+=1
        print("%d pairs"%(count))
        fp.close()
    scPairsSummary.close()

def awk():
    cmd = 'awk -F\'\\t\' \'{print $1 \"\\t\"  $3 \"_\" $4}\''
    cmd +=  ' '  'read_barcode.index > barcode.index' 
    print(cmd) 
    return run_cmd_return_time(cmd)

def samtools_merge(fastq):
    cmd = 'samtools merge *.bam -o ' + fastq + '.merged.bam -f'
    time1 = run_cmd_return_time(cmd)
    cmd = 'rm *.bam'
    time2 = run_cmd_return_time(cmd)
    cmd = 'mv ' + fastq + '.merged.bam ' + fastq + '.bam'
    time3 = run_cmd_return_time(cmd)
    return time1 + time2 + time3  

def Generate_snHIC_ReadsIndex():
    file1_path = 'trimmed-pair1.fastq.gz'
    file2_path = 'trimmed-pair2.fastq.gz'
    indexName = 'barcode.index'
      #####记录barcode对应的reads
    fp = open(indexName, "w")
    DirtyFile = open("DirtyReads.txt", "w")

    # 打开两个文件并同时读取内容
    with gzip.open(file1_path, 'rt') as file1, gzip.open(file2_path, 'rt') as file2:
        # 同时逐行读取两个文件的内容
        for line1, line2 in zip(file1, file2):
            # 处理文件内容
            # 例如，打印每一行
            line1 = str(line1)
            line2 = str(line2)
            if '@' == line1[0] and '@' == line2[0]:
                if line1.split(" ")[0] == line2.split(" ")[0]:
                    if line1.split(" ")[1][-18:-1] == line2.split(" ")[1][-18:-1]:
                        fp.write(line1.split(" ")[0][1:]+"\t"+line2.split(" ")[1][-18:-1]+"\n")
                    else:
                        DirtyFile.write("Reads1\t"+line1[0:-1]+"\tReads2\t"+line2)
                else:
                    DirtyFile.write("Reads1\t" + line1[0:-1] + "\tReads2\t" + line2)
    fp.close()
    DirtyFile.close()


def split(fn1, size1, size2):
    if 'gz' in fn1:
            dfh1=gzip.open(fn1,'r')

    if 'gz' not in fn1:
        dfh1=open(fn1,'r')
    if sys.version_info[0]==3:
        ID1=str(dfh1.readline().rstrip())[2:-1]
        line1=str(dfh1.readline().rstrip())[2:-1]
        plus1=str(dfh1.readline().rstrip())[2:-1]
        QS1=str(dfh1.readline().rstrip())[2:-1]
    if sys.version_info[0]==2:
        ID1=dfh1.readline().rstrip()
        line1=dfh1.readline().rstrip()
        plus1=dfh1.readline().rstrip()
        QS1=dfh1.readline().rstrip()
    rfhs=[]


    trim1=5
    trim2=5

    for i in range(1,4):
        rfhs.append(open(fn1+'_r'+str(i)+'.fq','w'))


    while line1:
        if len(line1[trim1:size1+trim1])>=30:
            rfhs[0].write(str(ID1)+'-1'+'\n'+str(line1[trim1:size1+trim1])+'\n'+str(plus1[0])+'\n'+str(QS1[trim1:size1+trim1])+'\n')
        if len(line1[trim1+size1:(-1*size2)-trim2])>=30:
            rfhs[0].write(str(ID1)+'-2'+'\n'+str(line1[trim1+size1:(-1*size2)-trim2])+'\n'+str(plus1[0])+'\n'+str(QS1[trim1+size1:(-1*size2)-trim2])+'\n')
        if len(line1[(-1*size2)-trim2:-1*trim2])>=30:
            rfhs[0].write(str(ID1)+'-3'+'\n'+str(line1[(-1*size2)-trim2:-1*trim2])+'\n'+str(plus1[0])+'\n'+str(QS1[(-1*size2)-trim2:-1*trim2])+'\n')
        if sys.version_info[0]==3:
            ID1=str(dfh1.readline().rstrip())[2:-1]
            line1=str(dfh1.readline().rstrip())[2:-1]
            plus1=str(dfh1.readline().rstrip())[2:-1]
            QS1=str(dfh1.readline().rstrip())[2:-1]
        if sys.version_info[0]==2:
            ID1=dfh1.readline().rstrip()
            line1=dfh1.readline().rstrip()
            plus1=dfh1.readline().rstrip()
            QS1=dfh1.readline().rstrip()  

    map(lambda x:x.close(),rfhs)

def addWildcards(string):
    foo= ""
    for i in range(len(string)):
            foo+=string[:i] + '.' + string[i+1:] + "|"
    return foo[:-1]

def rc(seq):

    rcd = ""
    rcs = {"A":"T","C":"G","T":"A","G":"C", "N":"N"}
    for char in seq[::-1]:
        rcd+=rcs[char]
    return rcd

def associateBarcodes(fhi_r1, fhi_r2, fho_r1, fho_r2, internal_bc):
    outfile = open('read_barcode.index', 'w')
    '''Takes a Fastq filehandle in and filehandle out for each read; writes barcode-clipped fastqs to\
    filehandle out and returns a dictionary with barcode pairs associated with given read names'''
    #fhi_r1 =  gzip.open(fhi_r1, "rt")
    #fhi_r2 = gzip.open(fhi_r2, "rt")
    seq_r1 = FastqGeneralIterator(fhi_r1)
    seq_r2 = FastqGeneralIterator(fhi_r2)
    seqzip = zip(seq_r1, seq_r2) #zip up the two fastq iterators to increase performance
    bc_seq = {}
    bc1s = [] #Container to store internal barcode IDs
    bridge = ''
    #Create a list of all possible internal barcodes (necessary
    #for hamming distance calculations)
    for line in internal_bc:
        bc = line.rstrip('\n')
        strip = 'GATC' + bc + 'GAATTC'
        bc1s.append(bc)
    bridge_re = re.compile('GATC[ACGT]{8}GAATTC')
    #We want to find all reads where
    #the internal adaptor is present
    #in either R1 or R2. Accepting that we're
    #throwing out some stuff because of mismatches,
    ##let's just use the re module to do this.
    for i in seqzip:
        title1, seq1, qual1 = i[0]
        title2, seq2, qual2 = i[1]
        new_title = title1.split()[0]
        m1 = bridge_re.search(seq1)
        m2 = bridge_re.search(seq2)
        terminal_bc = title1.split('&')[1]
        if m1:
            if m2:
                pos1 = m1.start()
                pos2 = m2.start()
                trim_seq1 = seq1[:pos1]
                trim_seq2 = seq2[:pos2]
                trim_qual1 = qual1[:pos1]
                trim_qual2 = qual2[:pos2]
                #Check for length of the read
                if len(trim_seq1) < 30 or len(trim_seq2) < 30:
                    continue #and continue if either one is too short
                barcode_seq1 = seq1[pos1 + 4:pos1 + 12]
                barcode_seq2 = seq2[pos2 + 4:pos2 + 12]
                ham_bc1 = checkHamming(bc1s, barcode_seq1)
                ham_bc2 = checkHamming(bc1s, barcode_seq2)
                if ham_bc1[0]:
                    if ham_bc2[0]:
                        print("@%s\n%s\n+\n%s" % (new_title, trim_seq1, trim_qual1), file=fho_r1)
                        print("@%s\n%s\n+\n%s" % (new_title, trim_seq2, trim_qual2), file=fho_r2)
                        bc1 = ham_bc1[1]
                        bc2 = ham_bc2[1]
                        print("%s\t%s\t%s\t%s" % (new_title, bc1, bc2, terminal_bc), file=outfile)
            else:
                pos1 = m1.start()
                pos1_end = m1.end()
                trim_seq1 = seq1[:pos1]
                trim_qual1 = qual1[:pos1]
                #Check for length of the rea
                if len(trim_seq1) < 30:
                    continue #and continue if either one is too short
                if len(seq1[pos1_end:]) < 12:
                    continue
                barcode_seq2 = rc(seq1[pos1_end:pos1_end + 8])
                barcode_seq1 = seq1[pos1 + 4:pos1 + 12]
                ham_bc1 = checkHamming(bc1s, barcode_seq1)
                ham_bc2 = checkHamming(bc1s, barcode_seq2)
                if ham_bc1[0]:
                    if ham_bc2[0]:
                        print("@%s\n%s\n+\n%s" % (new_title, trim_seq1, trim_qual1), file=fho_r1)
                        print("@%s\n%s\n+\n%s" % (new_title, seq2, qual2), file=fho_r2)
                        bc1 = ham_bc1[1]
                        bc2 = ham_bc2[1]
                        print("%s\t%s\t%s\t%s" % (new_title, bc1, bc2, terminal_bc),file=outfile)
        else:
            if m2:
                pos2 = m2.start()
                pos2_end = m2.end()
                trim_seq2 = seq2[:pos2]
                trim_qual2 = qual2[:pos2]
                #Check for length of the read
#				print seq2[pos2_end:]
                if len(trim_seq2) < 30:
                    continue #and continue if either one is too short
                if len(seq2[pos2_end:]) < 12:
                    continue #and continue if either one is too short
                barcode_seq1 = rc(seq2[pos2_end:pos2_end + 8])
                barcode_seq2 = seq2[pos2 + 4:pos2 + 12]
                ham_bc1 = checkHamming(bc1s, barcode_seq1)
                ham_bc2 = checkHamming(bc1s, barcode_seq2)
                if ham_bc1[0]:
                    if ham_bc2[0]:
                        print("@%s\n%s\n+\n%s" % (new_title, seq1, qual1), file=fho_r1)
                        print("@%s\n%s\n+\n%s" % (new_title, trim_seq2, trim_qual2), file=fho_r2)
                        bc1 = ham_bc1[1]
                        bc2 = ham_bc2[1]
                        print("%s\t%s\t%s\t%s" % (new_title, bc1, bc2, terminal_bc),file=outfile)

def analyze_scDHC_V2design(fhi_bc, fhi_r1,fhi_r2,fho_r1,fho_r2):
	fhi_bc = open(fhi_bc )
	# fhi_r1 = gzip.open(fhi_r1,'r')
	# fhi_r2 = gzip.open(fhi_r2,'r' )
	fho_r1 = open(fho_r1 , 'w')
	fho_r2 = open(fho_r2, 'w')
	assoc = associateBarcodes(fhi_r1, fhi_r2, fho_r1, fho_r2, fhi_bc)
	fhi_bc.close()
	# fhi_r1.close()
	# fhi_r2.close()
	fho_r1.close()
	fho_r2.close()
     
def checkHamming(barcodes,barcode):
	'''Given a list of barcodes, check that the given barcode is within edit\
		distance 2 to any of the list of barcodes'''
	for bc in barcodes:
		match = False
		hd = hamming(barcode, bc)
		if hd <= 2:
			match = True
			barcode = bc
			break
	return (match, barcode)



def inline_splitter(r1, r2, barcode_fhi, r1_o, r2_o, log):

	barcode_fhi = open(barcode_fhi) #Barcode filehandle
	#Barcodes
	barcodes = []
	for line in barcode_fhi:
		barcodes.append(line.split()[0])
	r1_o = open(r1_o, 'w') #R1 filehandle out
	r2_o = open(r2_o, 'w') #R2 filehandle out
	mismatch, not_found_R1, not_found_R2, reads = split_fastqs(r1, r2, r1_o, r2_o, barcodes) #Split some fastqs, yo
	#Close the filehandles
	# r1.close()
	# r2.close()
	r1_o.close()
	r2_o.close()
	barcode_fhi.close()

	# 将mismatch, not_found_R1, not_found_R2, reads写入log
	with open(log, 'w') as f:
		f.write('mismatch: ' + str(mismatch) + '\n')
		f.write('not_found_R1: ' + str(not_found_R1) + '\n')
		f.write('not_found_R2: ' + str(not_found_R2) + '\n')
		f.write('reads: ' + str(reads) + '\n')


	f.close()
     
def cload_correct_zoomify(dir_, pair, genomesize,res,resz):
    target = dir_ + pair
    name = pair.split('.')[0]
    cmd = 'cooler cload pairs -c1 2 -p1 3 -c2 4 -p2 5 '
    cmd += genomesize + ':' + str(res) + ' ' + target + ' ' + dir_  +  name + '_' + str(res) + '.cool'
    # print(cmd)
    time1 = run_cmd_return_time(cmd)

    # cmd = 'hicCorrectMatrix correct --matrix  ' + dir_ + name + '_' + str(res) + '.cool'
    # cmd += ' --correctionMethod KR --outFileName '
    # cmd += ' ' + dir_ + name + '_' + str(res) + '.KR.cool'
    # cmd += ' --filterThreshold -1.5 5.0 --chromosomes chr1 chr2 chr3 chr4 chr5 chr6 chr7 chr8 chr9 chr10 chr11 chr12 chr13 chr14 chr15 chr16 chr17 chr18 chr19'
    # # print(cmd)
    # time1 = run_cmd_return_time(cmd)


    cmd = 'cooler zoomify '  + \
         dir_ + name + '_' +str(res) + \
        '.cool -r ' + resz + ' -o ' + dir_ + name + '.mcool'
    time1 = run_cmd_return_time(cmd) 



def snHic_barcode(file1_path, file2_path):
    #####记录barcode对应的reads
    fp = open('barcode.index', "w")
    DirtyFile = open("DirtyReads.txt", "w")

    # 打开两个文件并同时读取内容
    with gzip.open(file1_path, 'rt') as file1, gzip.open(file2_path, 'rt') as file2:
        # 同时逐行读取两个文件的内容
        for line1, line2 in zip(file1, file2):
            # 处理文件内容
            # 例如，打印每一行
            line1 = str(line1)
            line2 = str(line2)
            if '@' == line1[0] and '@' == line2[0]:
                if line1.split(" ")[0] == line2.split(" ")[0]:
                    if line1.split(" ")[1][-18:-1] == line2.split(" ")[1][-18:-1]:
                        fp.write(line1.split(" ")[0][1:]+"\t"+line2.split(" ")[1][-18:-1]+"\n")
                    else:
                        DirtyFile.write("Reads1\t"+line1[0:-1]+"\tReads2\t"+line2)
                else:
                    DirtyFile.write("Reads1\t" + line1[0:-1] + "\tReads2\t" + line2)
    fp.close()
    DirtyFile.close()


import os
import matplotlib.pyplot as plt
import numpy as np

def DrawCellInternerCount(CellCounts = None, ChromLenList = None,save=''):
    # Fig.1 Cells Chromatain Interner Count and Proporation
    ###Get Chrom Length
    values_list = [value for value in ChromLenList[0].values()]
    dataPoints = sum(np.multiply(values_list, values_list))
    ###Get propotion
    CellCounts = sorted(CellCounts)
    CellProporty = CellCounts / dataPoints
    ###Log10 count
    CellCounts = np.log10(np.array(CellCounts) + 1)
    ###Draw Plot
    x = list(range(1, len(CellCounts)+1))
    # 创建图形和坐标轴对象
    fig, ax1 = plt.subplots()
    # 绘制第一个列表的数据
    ax1.plot(x, CellCounts, 'ro-', label='Count')
    ax1.set_xlabel('Cell Number')
    ax1.set_ylabel('1*10^n interactions')
    ax1.tick_params(axis='y')
    # 创建第二个纵轴并绘制第二个列表的数据
    ax2 = ax1.twinx()
    ax2.plot(x, CellProporty, 'bs--', label='Proporation')
    ax2.set_ylabel('Proportion in whole interactions')
    ax2.tick_params(axis='y')
    # 添加图例
    lines, labels = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax1.legend(lines + lines2, labels + labels2)
    # 设置标题
    plt.title('Fig.1 Cells Chromatain Interner Count and Proporation')
    # 显示图形
    plt.savefig(save + '/Fig.1 Cells Chromatain Interner Count and Proporation.png',dpi=300)
    plt.close()


def DrawChromatainInternerCount(ChromIntraction = None, ChromLenList = None, save=''):

    ###Get Chrom Name
    ChromsName = set()
    for term in list(ChromIntraction[0].keys()):
        ChromsName.add(term.split("-")[0])
        ChromsName.add(term.split("-")[1])
    ChromsName = list(ChromsName)

    InteractionName = list(ChromIntraction[0].keys())

    InteractionArr = []
    for i in range(len(ChromIntraction)):
        InteractionArr.append([value for value in ChromIntraction[i].values()])
    InteractionArr = np.array(InteractionArr)

    ChromInner = []
    ChromOuter = [[] for _ in range(len(ChromsName))]

    for i in range(len(ChromsName)):
        for j in range(len(ChromsName)):

            if ChromsName[i] == ChromsName[j]:
                ChromInner.append(ChromsName[i]+"-"+ChromsName[j])
            else:
                ChromOuter[i].append(ChromsName[i]+"-"+ChromsName[j])
    ####inner
    ChromInnerpositions = []
    ChromInnerName = []
    for term in ChromInner:
        ChromInnerpositions.append(InteractionName.index(term))
        ChromInnerName.append(term.split("-")[0])
    ChromInnerInteraction = InteractionArr[:, ChromInnerpositions]

    ###Outner
    ChromOuterInteraction = []
    CellouterInteraction = []
    ChromOuterName = []
    for i in range(len(ChromsName)):
        curChromouterpositions = []
        ChromOuterName.append(ChromOuter[i][0].split("-")[0])

        for term in ChromOuter[i]:
            if term not in InteractionName:
                term = term.split("-")[1]+"-"+term.split("-")[0]
            curChromouterpositions.append(InteractionName.index(term))
        curChromOuterInteraction = InteractionArr[:, curChromouterpositions]
        CellouterInteraction.append(curChromOuterInteraction)
        curChromOuterInteraction = curChromOuterInteraction.flatten()
        ChromOuterInteraction.append(curChromOuterInteraction)


    #### Fig.2 Draw CellouterInteraction
    CellouterInteraction = np.hstack(CellouterInteraction)  #### CellouterInteraction
    CellouterInteraction = np.log10(CellouterInteraction + 1)
    # 计算每个细胞的平均值和标准差
    mean_values = np.mean(CellouterInteraction, axis=1)
    std_values = np.std(CellouterInteraction, axis=1)
    # sort
    mean_values = mean_values[np.argsort(mean_values)]
    std_values = std_values[np.argsort(mean_values)]
    # 创建一个图形对象和一个子图对象
    fig, ax = plt.subplots()
    # 绘制平均值线
    ax.errorbar(np.arange(len(mean_values)), mean_values, yerr=std_values, fmt='o', capsize=5)
    # 设置图形的标题和坐标轴标签
    ax.set_title('Fig.2 Cell Chromatain Outer Interactions')
    ax.set_xlabel('Cell number')
    ax.set_ylabel('1*10^N interactions')
    ax.set_ylim(bottom=0)
    # 显示图形
    plt.savefig(save + '/Fig.2 Cell Chromatain Outer Interactions.png',dpi=300)
    plt.close()


    #### Fig.3 Draw Chromatain Inner Interaction
    ChromInnerInteraction = np.log10(ChromInnerInteraction + 1)
    # 计算每个细胞的平均值和标准差
    mean_values = np.mean(ChromInnerInteraction, axis=0)
    std_values = np.std(ChromInnerInteraction, axis=0)
    # sort
    mean_values = mean_values[np.argsort(ChromsName)]
    std_values = std_values[np.argsort(ChromsName)]
    ChromsName_inner = [ChromsName[i] for i in np.argsort(ChromsName)] # 使用排序后的索引来获取对应的横坐标名字
    fig, ax = plt.subplots(figsize=(12, 12 / 1.618 ))
    # 绘制平均值线
    ax.errorbar(np.arange(len(mean_values)), mean_values, yerr=std_values, fmt='o', capsize=5)
    # 设置图形的标题和坐标轴标签
    x_ticks = np.arange(len(mean_values))
    x_tick_labels = ChromsName_inner
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(x_tick_labels)
    ax.set_title('Fig.3 Cell Chromatain Inner Interactions')
    ax.set_xlabel('Chromatain Names')
    ax.set_ylabel('1*10^N interactions')
    ax.set_ylim(bottom=0)
    # 显示图形
    plt.savefig(save + '/Fig.3 Cell Chromatain Inner Interactions.png',dpi=300)
    plt.close()

    #### Fig.4 Draw Chromatain Outer Interaction
    ChromOuterInteraction = np.array(ChromOuterInteraction)
    # 计算每个染色体的平均值和标准差
    mean_values = np.mean(ChromOuterInteraction, axis=1)
    std_values = np.std(ChromOuterInteraction, axis=1)
    # sort
    mean_values = mean_values[np.argsort(ChromsName)]
    std_values = std_values[np.argsort(ChromsName)]
    ChromsName_outer = [ChromsName[i] for i in np.argsort(ChromsName)] # 使用排序后的索引来获取对应的横坐标名字
    # 创建一个图形对象和一个子图对象
    fig, ax = plt.subplots(figsize=(12, 12 / 1.618 ))
    # 绘制平均值线
    ax.errorbar(np.arange(len(mean_values)), mean_values, yerr=std_values, fmt='o', capsize=5)
    # 设置图形的标题和坐标轴标签
    x_tick_labels = ChromsName_outer
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(x_tick_labels)
    ax.set_title('Fig.4 Chromatain Outer Interactions')
    ax.set_xlabel('Cell number')
    ax.set_ylabel('1*10^N interactions')
    ax.set_ylim(bottom=0)
    # 显示图形
    plt.savefig(save + '/Fig.4 Chromatain Outer Interactions.png',dpi=300)
    plt.close()

def DrawDistanceCellInternerCount(CellCounts=None, ChromLenList=None, save=None):
    # Fig.5 Cells Chromatain Interner distance interacions and Proporation
    ###Get Chrom Length
    values_list = [value for value in ChromLenList[0].values()]
    dataPoints = sum(np.multiply(values_list, values_list))
    ###Get propotion
    CellCounts = sorted(CellCounts)
    CellProporty = CellCounts / dataPoints
    ###Log10 count
    CellCounts = np.log10(np.array(CellCounts) + 1)
    ###Draw Plot
    x = list(range(1, len(CellCounts) + 1))
    # 创建图形和坐标轴对象
    fig, ax1 = plt.subplots()
    # 绘制第一个列表的数据
    ax1.plot(x, CellCounts, 'ro-', label='Count')
    ax1.set_xlabel('Cell Number')
    ax1.set_ylabel('1*10^n interactions')
    ax1.tick_params(axis='y')
    # 创建第二个纵轴并绘制第二个列表的数据
    ax2 = ax1.twinx()
    ax2.plot(x, CellProporty, 'bs--', label='Proporation')
    ax2.set_ylabel('Proportion in whole interactions')
    ax2.tick_params(axis='y')
    # 添加图例
    lines, labels = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax1.legend(lines + lines2, labels + labels2)
    # 设置标题
    plt.title('Fig.5 Cells Chromatain Interner distance interacions and Proporation')
    # 显示图形
    plt.savefig(save + '/Fig.5 Cells Chromatain Interner distance interacions and Proporation.png', dpi=300)
    plt.close()


def DrawDistanceChromatainInternerCount(ChromatainCounts=None, save=None):
    # Fig.6 Chromatain distance Interactions
    InteractionArr = []
    for i in range(len(ChromatainCounts)):
        InteractionArr.append([value for value in ChromatainCounts[i].values()])
    InteractionArr = np.array(InteractionArr)
    ChromsName = list(ChromatainCounts[0].keys())

    ChromOuterInteraction = np.array(InteractionArr)
    # 计算每个染色体的平均值和标准差
    mean_values = np.mean(ChromOuterInteraction, axis=0)
    std_values = np.std(ChromOuterInteraction, axis=0)
    # sort
    mean_values = mean_values[np.argsort(ChromsName)]
    std_values = std_values[np.argsort(ChromsName)]
    ChromsNames = [ChromsName[i] for i in np.argsort(ChromsName)]  # 使用排序后的索引来获取对应的横坐标名字
    # 创建一个图形对象和一个子图对象
    fig, ax = plt.subplots(figsize=(12, 12 / 1.618))
    # 绘制平均值线
    ax.errorbar(np.arange(len(mean_values)), mean_values, yerr=std_values, fmt='o', capsize=5)
    # 设置图形的标题和坐标轴标签
    x_ticks = np.arange(len(mean_values))
    x_tick_labels = ChromsNames
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(x_tick_labels)
    ax.set_title('Fig.6 Chromatain distance Interactions')
    ax.set_xlabel('Cell number')
    ax.set_ylabel('1*10^N interactions')
    ax.set_ylim(bottom=0)
    # 显示图形
    plt.savefig(save + '/Fig.6 Chromatain distance Interactions.png', dpi=300)
    plt.close()






















import os
import cooler
import cooltools
import numpy as np
import pandas as pd
from cooltools import insulation
import bioframe

import statsmodels.api as sm
# linear
from sklearn.linear_model import Ridge
from joblib import Parallel, delayed
import click

pre_chrom = ['chr' + str((i + 1)) for i in range(22)] + ['chrX', 'chrY']


def rg(prepare):
    prepare = prepare.loc[~(prepare == 0).all(axis=1)]
    prepare = np.log2(prepare + 1)
    X = prepare[["dup", "trans", "tad", "ab"]]
    y = prepare["contacts"]
    # 使用statsmodels进行线性回归
    # 使用statsmodels进行线性回归
    X = sm.add_constant(X)

    # 使用岭回归进行拟合
    ridge = Ridge(alpha=1.0, fit_intercept=False)  # 设置fit_intercept=False,让Ridge不再估计截距项
    ridge.fit(X, y)

    # 提取标准化系数的绝对值
    beta_abs = np.abs(ridge.coef_)

    # 根据变量的相关性分为正负两组,并将常数项视为负相关指标
    pos_vars = ["tad", "ab"]
    neg_vars = ["dup", "trans", "const"]

    # 计算正负变量标准化系数的和
    pos_sum = beta_abs[X.columns.isin(pos_vars)].sum()
    neg_sum = beta_abs[X.columns.isin(neg_vars)].sum()

    # 计算有效信息指标
    valid_info = pos_sum / neg_sum

    return valid_info




def rg(prepare):
    prepare = prepare.loc[~(prepare == 0).all(axis=1)]
    prepare = np.log2(prepare + 1)
    X = prepare[["trans", "tad", "ab"]]
    y = prepare["contacts"]
    # 使用statsmodels进行线性回归
    # 使用statsmodels进行线性回归
    X = sm.add_constant(X)

    # 使用岭回归进行拟合
    ridge = Ridge(alpha=1.0, fit_intercept=False)  # 设置fit_intercept=False,让Ridge不再估计截距项
    ridge.fit(X, y)

    # 提取标准化系数的绝对值
    beta_abs = np.abs(ridge.coef_)

    # 根据变量的相关性分为正负两组,并将常数项视为负相关指标
    pos_vars = ["tad", "ab"]
    neg_vars = ["trans", "const"]

    # 计算正负变量标准化系数的和
    pos_sum = beta_abs[X.columns.isin(pos_vars)].sum()
    neg_sum = beta_abs[X.columns.isin(neg_vars)].sum()

    # 计算有效信息指标
    valid_info = pos_sum / neg_sum

    return valid_info


def deal(mcool, sp, g, name, output):
    cache_path = './cache'
    os.makedirs(cache_path, exist_ok=True)
    if os.path.exists(output + '/' + name + '.txt'):
        return
    print(mcool)

    try:
        clr = cooler.Cooler('%s::/resolutions/40000' % mcool)
    except:
        print('err load, re-zoomify')
        cooler.zoomify_cooler(base_uris='%s::/resolutions/10000' % mcool,
                              outfile=cache_path + '/' + os.path.basename(mcool),
                              resolutions=[10000, 40000, 500000, 1000000],
                              chunksize=100000000, nproc=10)
        os.system('mv %s %s' % (cache_path + '/' + os.path.basename(mcool), mcool))
        clr = cooler.Cooler('%s::/resolutions/40000' % mcool)
    if not cooltools.lib.is_cooler_balanced(clr):
        cooler.balance_cooler(clr, store=True)
    win = [10 * 40000]
    insulation_table = insulation(clr, win, verbose=False)
    insulation_table = insulation_table.groupby('chrom')['is_boundary_400000'].sum()
    insulation_table = pd.DataFrame(insulation_table[insulation_table.index.isin(pre_chrom)])
    insulation_table.columns = ['tad']
    print('done IS')
    clr = cooler.Cooler('%s::/resolutions/1000000' % mcool)
    if not cooltools.lib.is_cooler_balanced(clr):
        cooler.balance_cooler(clr, store=True)
    genome = bioframe.load_fasta(g)
    bins = clr.bins()[:]
    gc_cov = bioframe.frac_gc(bins[['chrom', 'start', 'end']], genome)
    view_df = pd.DataFrame({'chrom': clr.chromnames,
                            'start': 0,
                            'end': clr.chromsizes.values,
                            'name': clr.chromnames}
                           )

    try:
        cis_eigs = cooltools.eigs_cis(
            clr,
            gc_cov,
            view_df=view_df,
            n_eigs=3,
        )

        # cis_eigs[0] returns eigenvalues, here we focus on eigenvectors
        eigenvector_track = cis_eigs[1][['chrom', 'start', 'end', 'E1']]

        eigenvector_track['E1'] = np.nan_to_num(eigenvector_track['E1'])
        eigenvector_track['sign'] = '$'
        eigenvector_track['sign'][eigenvector_track['E1'] > 0] = 'A'
        eigenvector_track['sign'][eigenvector_track['E1'] < 0] = 'B'
        a_list = []
        b_list = []
        ab_switch = {}
        for c in pre_chrom:
            ab_switch[c] = 0
        for c in pre_chrom:
            if not c in set(eigenvector_track['chrom']):
                continue
            seq = list(eigenvector_track[eigenvector_track['chrom'] == c]['sign'])
            base = eigenvector_track[eigenvector_track['chrom'] == c].index[0]

            seq = ''.join(seq)
            #     print(seq)
            new_seq = ''
            idx = []
            tmp_idx = []
            for i in range(len(seq)):
                sym = seq[i]
                if new_seq == '':
                    new_seq += sym
                    tmp_idx.append(i + base)
                else:
                    if sym == seq[i - 1]:
                        tmp_idx.append(i + base)
                    else:
                        idx.append(tmp_idx)
                        tmp_idx = [i + base]
                        new_seq += sym
            for i in range(len(new_seq) - 3):
                if new_seq[i:i + 3] == 'ABA':
                    tmp_A = idx[i] + idx[i + 2]
                    tmp_B = idx[i + 1]
                    ab_switch[c] += 1

                elif new_seq[i:i + 3] == 'BAB':
                    tmp_A = idx[i + 1]
                    tmp_B = idx[i] + idx[i + 2]
                    ab_switch[c] += 1
        ab_switch = pd.DataFrame(ab_switch, index=[0]).T
        ab_switch.columns = ['ab']
        print(ab_switch)
    except:
        ab_switch = pd.DataFrame({c: 0 for c in pre_chrom}, index=[0]).T
        ab_switch.columns = ['ab']

    clr = cooler.Cooler('%s::/resolutions/40000' % mcool)
    bins = clr.bins()[:]
    pixels = clr.pixels()[:]
    pixels['chr1'] = list(bins.iloc[list(pixels['bin1_id']), 0])
    pixels['chr2'] = list(bins.iloc[list(pixels['bin2_id']), 0])
    pixels['start'] = list(bins.iloc[list(pixels['bin1_id']), 1])
    pixels['end'] = list(bins.iloc[list(pixels['bin1_id']), 2])
    pixels = pixels[pixels['chr1'].isin(pre_chrom) & pixels['chr2'].isin(pre_chrom)]
    trans_result = pixels[pixels['chr1'] != pixels['chr2']]
    trans1 = pd.DataFrame(trans_result.groupby('chr1')['count'].sum())
    trans2 = pd.DataFrame(trans_result.groupby('chr2')['count'].sum())
    trans_result = pd.concat([trans1, trans2], axis=1)
    trans_result.fillna(0, inplace=True)
    trans_result = pd.DataFrame(trans_result.sum(1))
    trans_result.columns = ['trans']
    intra = pixels[pixels['chr1'] == pixels['chr2']]
    intra = pd.DataFrame(intra.groupby('chr1')['count'].sum())

    raw = {}
    for c in set(pixels['chr1']):
        raw[c] = pixels[(pixels['chr1'] == c) | (pixels['chr2'] == c)]['count'].sum()

    raw = pd.DataFrame(raw, index=[0]).T
    raw.columns = ['contacts']

    result = pd.concat([raw, trans_result, ab_switch, insulation_table], axis=1)
    result.fillna(0, inplace=True)
    result.to_csv(output + '/' + name + '.csv')
    result = pd.read_csv(output + '/' + name + '.csv')
    result = result.set_index('Unnamed: 0')
    try:
        valid = rg(result)
    except:
        valid = 0

    with open(output + '/' + name + '.txt', 'w+') as f:
        f.write(str(valid))
    f.close()

    return valid



@click.command()
@click.option('--mcool', help='mcool file path')
@click.option('--genome', help='genome file path')
@click.option('--output', help='save path')
@click.option('--nproc', help='number of process')
def ssce(mcool, g,  output, nproc=10):
    task = []
    for mcool in os.listdir(mcool):
        if mcool.endswith('.mcool'):
            name = mcool.split('.')[0]
            task.append((mcool, g, name, output))
    Parallel(n_jobs=nproc)(delayed(deal)(mcool, sp, g, name, output) for mcool, sp, g, name, output in task)



#
# This script can be run using the following command:
#  
#    python setup_cython.py build_ext --inplace
#

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
from numpy import get_include as numpy_get_include

numpy_includes = numpy_get_include()

ext_modules = [ Extension("sc3dg.model.dyn_util", ["sc3dg/model/dyn_util.pyx"], libraries=["m"], include_dirs=[numpy_includes]) ]

setup(name='NucDynamics Cython utilities',  cmdclass={'build_ext':build_ext},  ext_modules=ext_modules)



import argparse
import pysam



def reformat_bam(input_sam, output_sam, input_is_bam, output_is_bam, set_bx_flag):

    read_indices = {}
    for align_idx, align in enumerate(input_sam.fetch(until_eof=True)):
        read_id = align.query_name
        read_idx = read_indices.get(read_id, None)
        if read_idx is None:
            read_idx = len(read_indices)
            read_indices[read_id] = read_idx
        if set_bx_flag:
            align.set_tag(tag="BX", value=align.query_name, value_type="Z")
        align.query_name = f"{read_id}:{read_idx}:{align_idx}"
        output_sam.write(align)
    output_sam.close()






def reSAM(input_sam, output_sam, input_is_bam, output_is_bam, set_bx_flag):


    # 打开输入和输出文件
    input_format = "rb" if input_is_bam else "r"
    output_format = "wb" if output_is_bam else "w"

    with pysam.AlignmentFile(input_sam, input_format) as input_sam:
        # 获取输入文件的头部信息，用于初始化输出文件
        header = input_sam.header

        # 打开输出文件，并传递头部信息来初始化它
        with pysam.AlignmentFile(output_sam, output_format, header=header) as output_sam:
            # 调用函数进行重新格式化
            reformat_bam(input_sam, output_sam, input_is_bam, output_is_bam, set_bx_flag)



from sc3dg.analysis.api import MMCooler
import click
from joblib import Parallel, delayed
import cooler
import os
import cooltools
import pandas as pd
from cooltools import insulation
chrom = ['chr' + str(i) for i in range(20)]
@click.command('accum')
@click.option('--mcool',  help='mcool path',required=True, default=None)
@click.option('--output',  help='output path',required=True, default=None)
@click.option('--resolution', help='resolution',required=True, default=40000, type=int)
@click.option('--top_num', help='top number',required=True, default=5, type=int)
def accum(mcool, output, resolution, top_num, describe='accum'):
    pass
    ob1 = MMCooler(directory=mcool,
                   resolution=[resolution],
                   describe = describe,
                   merge=False
                   )
    ob1.sorted_cell(num=top_num)

    iou0 = []  # for different windowns
    iou1 = []
    iou2 = []
    iou3 = []
    iou = []
    win = [3 * resolution, 5 * resolution, 10 * resolution, 25 * resolution]
    print('merging')
    cache_file = []
    for clr in ob1.mcool_files:
        cache_file.append(clr + '::resolutions/%s' % resolution)

    if not os.path.exists('./cache/%s_assemb.mcool' % (describe)):
        print('merging')
        cooler.merge_coolers(input_uris=cache_file, output_uri='./cache/%s_assemb.mcool' % (describe),
                             mergebuf=10000000000)

    clr = cooler.Cooler('./cache/%s_assemb.mcool' % (describe))
    if not cooltools.lib.is_cooler_balanced(clr):
        cooler.balance_cooler(clr, mad_max=0, min_nnz=0, ignore_diags=1
                              , store=True, store_name='weight')

    insulation_table = insulation(clr, win, verbose=True, nproc=20)
    boundary = insulation_table[insulation_table['is_boundary_' + str(win[0])] == True]
    boundary2 = boundary[boundary['chrom'].isin(chrom)]

    print(boundary2)

    num = len(ob1.clr_obj[0])
    tmp0 = []
    tmp1 = []
    tmp2 = []
    tmp3 = []
    tmp = []
    count0 = []
    count1 = []
    count2 = []
    count3 = []



    ob1.pooled_cell(pool_size=1000, num=min(num, top_num))

    task = []
    for i, clr in enumerate(ob1.clr_obj[-1]):
        print(i)
        task.append((clr, i))

    def balance(clr, c):
        print(clr, c)
        if not cooltools.lib.is_cooler_balanced(clr):
            cooler.balance_cooler(clr, mad_max=0, min_nnz=0, ignore_diags=1
                                  , store=True, store_name='weight')

    Parallel(n_jobs=10)(delayed(balance)(clr, c) for clr, c in task)

    for i, clr in enumerate(ob1.clr_obj[-1]):
        print(i)
        insulation_table = insulation(clr, win, verbose=False, )
        boundary = insulation_table[insulation_table['is_boundary_' + str(win[0])] == True]
        boundary = boundary[boundary['chrom'].isin(chrom)]
        b0 = boundary[boundary['is_boundary_' + str(win[0])] == True]
        b1 = boundary[boundary['is_boundary_' + str(win[1])] == True]
        b2 = boundary[boundary['is_boundary_' + str(win[2])] == True]
        b3 = boundary[boundary['is_boundary_' + str(win[3])] == True]

        b00 = boundary2[boundary2['is_boundary_' + str(win[0])] == True]
        b11 = boundary2[boundary2['is_boundary_' + str(win[1])] == True]
        b22 = boundary2[boundary2['is_boundary_' + str(win[2])] == True]
        b33 = boundary2[boundary2['is_boundary_' + str(win[3])] == True]

        iou_0 = len(set(b0.index).intersection(set(b00.index))) / len(set(b0.index).union(set(b00.index)))
        iou_1 = len(set(b1.index).intersection(set(b11.index))) / len(set(b1.index).union(set(b11.index)))
        iou_2 = len(set(b2.index).intersection(set(b22.index))) / len(set(b2.index).union(set(b22.index)))
        iou_3 = len(set(b3.index).intersection(set(b33.index))) / len(set(b3.index).union(set(b33.index)))
        iou_ = len(set(boundary.index).intersection(set(boundary2.index))) / len(
            set(boundary.index).union(set(boundary2.index)))

        tmp0.append(iou_0)
        tmp1.append(iou_1)
        tmp2.append(iou_2)
        tmp3.append(iou_3)
        tmp.append(iou_)

        count0.append(b0.shape[0])
        count1.append(b1.shape[0])
        count2.append(b2.shape[0])
        count3.append(b3.shape[0])

    iou0.append(tmp0)
    iou1.append(tmp1)
    iou2.append(tmp2)
    iou3.append(tmp3)
    iou.append(tmp)

    res = pd.DataFrame(iou0)
    res.to_csv('%s/bulkvspooled_0_%s.csv' % (output, describe), index=False)
    res = pd.DataFrame(iou1)
    res.to_csv('%s/bulkvspooled_1_%s.csv' % (output, describe), index=False)
    res = pd.DataFrame(iou2)
    res.to_csv('%s/bulkvspooled_2_%s.csv' % (output, describe), index=False)
    res = pd.DataFrame(iou3)
    res.to_csv('%s/bulkvspooled_3_%s.csv' % (output, describe), index=False)


    res = pd.DataFrame(count0)
    res.to_csv('%s/bulkvspooled_0_%s_count.csv' % (output, describe), index=False)
    res = pd.DataFrame(count1)
    res.to_csv('%s/bulkvspooled_1_%s_count.csv' % (output, describe), index=False)
    res = pd.DataFrame(count2)
    res.to_csv('%s/bulkvspooled_2_%s_count.csv' % (output, describe), index=False)
    res = pd.DataFrame(count3)
    res.to_csv('%s/bulkvspooled_3_%s_count.csv' % (output, describe), index=False)
    res = pd.DataFrame(iou)
    res.to_csv('%s/bulkvspooled_%s.csv' % (output, describe), index=False)







import os
import gzip
import sys
import argparse
import re
import numpy as np
import pandas as pd
import subprocess
import time
import logging
from pathlib import Path
from .scNano_barcode import generate_barcode
from .ReformSAM import reSAM
from .Generate_scNanoHIC_pairs import WriteCellPairs
from multiprocessing import Pool
barcodesh = '/cluster/home/Kangwen/Hic/ckw/utils/barcode.sh'

def run_command(command):
    print(command)
    subprocess.run(command, shell=True, timeout=None)
    

def pp(opt, fastq,log_out):
    print('==========run scNano================')

 
    T = time.time()
    # 移动到工作目录
    os.chdir(opt['output'])

    # 创建文件夹tmp,并移动
    if not os.path.exists(opt['type'] + '_' + fastq + '_tmp'): # scHic_sample_tmp
        os.makedirs(opt['type'] + '_' + fastq + '_tmp')
    output = opt['output'] + '/' +opt['type'] + '_' + fastq + '_tmp'
    os.chdir(opt['type'] + '_' + fastq + '_tmp')

    fq = opt['fastq_dir'][opt['fastq_log'].index(fastq)]
    PCR = os.path.dirname(fq) + '/' + 'PCR.txt'
    TN5 = os.path.dirname(fq) + '/' + 'TN5.txt'
    bc = os.path.dirname(fq) + '/' + 'index.txt'



    # 创建log
    logging.basicConfig(filename=fastq + '_logging.log', level=logging.DEBUG)
    for k in opt.keys():

     
        if (k == 'fastq_dir') or (k == 'fastq_log') or (k == 'run_sample'):
            continue
        else:
            logging.debug('# ' + k + ': ' + str(opt[k]))

    
    
    # 设置文件夹路径
    raw_folder = './raw_data'
    trim_folder = './trim_data'
    pairs_folder = './pairs_data'
    mapping_folder = './mapping_data'
    digestpairs_folder = './digestpairs_folder'
    filtedpairs_folder = './filted_pairs_data'
    sortedpairs_folder = './sorted_pairs_data'
    cool_folder = './cool_data'
    KRcool_folder = './KRcool_data'
    mcool_folder = './Mcool_data'

    # 创建必要的文件夹
    os.makedirs(trim_folder, exist_ok=True)
    os.makedirs(raw_folder, exist_ok=True)
    os.makedirs(pairs_folder, exist_ok=True)
    os.makedirs(mapping_folder, exist_ok=True)
    os.makedirs(digestpairs_folder, exist_ok=True)
    os.makedirs(filtedpairs_folder, exist_ok=True)
    os.makedirs(sortedpairs_folder, exist_ok=True)
    os.makedirs(cool_folder, exist_ok=True)
    os.makedirs(KRcool_folder, exist_ok=True)
    os.makedirs(mcool_folder, exist_ok=True)

    for val in os.listdir(opt['output'] + '/' + opt['type'] + '_' + fastq + '_tmp/Mcool_data'):
        if val.endswith('.mcool'):
            log_out.info('The result of ' + fastq + ' has been generated')
            print('The result of ' + fastq + ' has been generated')
            return

    with open(bc , 'r') as f:
        lines = f.readlines()
    lines = [x.strip() for x  in lines]

    divide_list = []
    bcname = []
    for i in range(0, len(lines), 3):
        divide_list.append(lines[i:i + 3])
        bcname.append('barcode_%s.fa' % i)


    for i,out in enumerate(bcname):
        print(out, divide_list[i])
        generate_barcode(divide_list[i],PCR, TN5,out)
    
    for fa in bcname:
        node = time.time()
        print(fa)
        cmd = 'nanoplexer -b %s -B 200M -t %s -p %s %s' % (fa, opt['thread'],raw_folder, fq)
        # run_command(cmd)
        print('nanoplexer:%s'% (time.time() - node))

    
    # 第一步: 运行 cutadapt 处理所有文件
    for file_name in os.listdir(raw_folder):
        if file_name.endswith(".fastq"):
            input_file = os.path.join(raw_folder, file_name)
            output_file = os.path.join(trim_folder, f"{file_name}_Trim.fastq")
            log_file = os.path.join(trim_folder, f"{file_name}_Trim.log")
            cutadapt_command = f"cutadapt -g AATGATACGGCGACCACCGAGATCT -g TCGTCGGCAGCGTC -g AGATGTGTATAAGAGACAG -a AGATCTCGGTGGTCGCCGTATCATT -a GACGCTGCCGACGA -a CTGTCTCTTATACACATCT --times 2 -j {opt['thread']} --minimum-length 500 -e 0.2 -O 12 -o {output_file} {input_file} > {log_file}"
            run_command(cutadapt_command)
    #         break

    
    for input_file in os.listdir(trim_folder):
        if input_file.endswith('.fastq'):
            # 提取文件名（不包括路径和扩展名）
            file_name = os.path.splitext(input_file)[0]
            
            # 设置输出文件的路径
            output_sam = os.path.join(mapping_folder, f'{file_name}.sam')
            reformed_sam = os.path.join(mapping_folder, f'{file_name}_reformed.sam')

            # 运行miniMap2命令
            reference_genome = opt['index']
            print('minimap2')
            t_=opt['thread']
            minimap2_cmd = f'minimap2 -ax map-ont -t {t_} {reference_genome} {os.path.join(trim_folder, input_file)} > {output_sam}'
            subprocess.run(minimap2_cmd, shell=True)

            # 运行Python脚本
            print('reSAM')
            reSAM(input_sam=output_sam,
                   output_sam=reformed_sam,
                   input_is_bam=False,
                     output_is_bam=False, 
                     set_bx_flag=False)

            WriteCellPairs(opt['genomesize'], reformed_sam,opt['species'],pairs_folder)
       

    # 查找指定目录中的所有 .pairs 文件
  
    for input_file in os.listdir(pairs_folder):
        if input_file.endswith(".pairs"):
            print(input_file)
            input_file_path = os.path.join(pairs_folder, input_file)
            file_name = os.path.splitext(input_file)[0].replace("_Trim_reformed", "")
            print(file_name)
            # 定义文件路径
            digest_pairs = os.path.join(digestpairs_folder, f"{file_name}_digest.pairs")
            filted_pairs = os.path.join(filtedpairs_folder, f"{file_name}_filted.pairs")
            sorted_pairs = os.path.join(sortedpairs_folder, f"{file_name}_sorted.pairs")

            # 执行 pairtools 命令
            print('restricting....')
            subprocess.run(["pairtools", "restrict", "-f", opt['enzyme_bed'], input_file_path, "-o", digest_pairs])
            print('selecting......')
            subprocess.run(["pairtools", "select", "mapq1>=30 and mapq2>=30", digest_pairs, "-o", filted_pairs])
            print('sorting.....')
            subprocess.run(["pairtools", "sort", "--nproc", str(opt['thread']), filted_pairs, "-o", sorted_pairs])

            resolution = int(opt['resolution'])
            cool_file = os.path.join(cool_folder, f"{file_name}.{resolution}.cool")
            KR_cool_file = os.path.join(KRcool_folder, f"{file_name}.{resolution}.KR.cool")
            mcool_file = os.path.join(mcool_folder, f"{file_name}.mcool")
            # 执行 cooler 和 hicCorrectMatrix 命令
            chrom_size = opt['genomesize']
            subprocess.run(["cooler", "cload", "pairs", "-c1", "2", "-p1", "3", "-c2", "4", "-p2", "5", f"{chrom_size}:{resolution}", sorted_pairs, cool_file])
            subprocess.run(["hicCorrectMatrix", "correct", "--matrix", cool_file, "--correctionMethod", "KR", "--outFileName", KR_cool_file, "--filterThreshold", "-1.5", "5.0", "--chromosomes", "chr1", "chr2", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9", "chr10", "chr11", "chr12", "chr13", "chr14", "chr15", "chr16", "chr17", "chr18", "chr19"])

            subprocess.run(['cooler', 'zoomify', KR_cool_file,'-r', opt['zoomify_res'], '-o', mcool_file])


import click
from sc3dg.analysis.api import MMCooler
import pandas as pd

@click.command('nDS')
@click.option('--mcool',  help='mcool path',required=True, default=None)
@click.option('--output',  help='output path',required=True, default=None)
@click.option('--resolution', help='resolution',required=True,  type=int)
@click.option('--mode', help='mode',required=True,type=click.Choice(['terr', 'tad', 'ab', 'cen']))
@click.option('--top_num', help='top number',required=False, default=None, type=int)
@click.option('--describe', help='describe',required=False, default='nDS')
@click.option('--epoch', help='epoch',required=False, default=10, type=int)
@click.option('--genome', help='fa genome',required=False, default='mm10')

def nDS(mcool, output, resolution, mode, top_num, describe, epoch, genome):

    '''

    Parameters
    ----------
    mcool:str
        mcool file path
    output:
        output file path
    resolution
    mode:str
        mode = ['terr', 'tad', 'ab', 'cen']
    top_num:int
        top number
    describe:str
        describe, for save file name
    epoch:int
        epoch
    Returns
    -------

    '''


    ob1 = MMCooler(directory=mcool,
                   resolution=[resolution],
                   describe = mode if not describe else describe,
                   merge=False
                   )

    if top_num:
        ob1.sorted_cell(num=top_num)

    

    if mode == 'terr':
        terr = ob1.data.get_nDS_chrom_terr(epoch=epoch,res=resolution)[0]
        idx = ob1.mcool_names
        terr.index = idx
        terr.to_csv(output, sep='\t')
    elif mode == 'tad':
        tad = ob1.get_TAD_nDS(epoch=epoch, res=resolution)
        df = pd.DataFrame({'tad':tad})
        idx = ob1.mcool_names
        df.index = idx
        df.to_csv(output, sep='\t')
    elif mode == 'ab':
        if not genome:
            raise ValueError('genome is required')
        ab = ob1.get_AB_compartment(res=resolution, epoch =epoch,genome=genome)[0]
        idx = ob1.mcool_names
        df = pd.DataFrame({'ab':ab})
        df.index = idx
        df.to_csv(output, sep='\t')
    elif mode == 'cen':
        cen = ob1.get_Centromere_nDs(epoch=epoch, res=resolution)
        idx = ob1.mcool_names
        cen.index = idx
        cen.to_csv(output, sep='\t')








import os
import glob
import cooler
import numpy as np
from scipy.stats import norm
from .embedding import *
import pandas as pd

class MultiMcoolReader:
    def __init__(self,directory, resolution):
        self.directory = directory
        self.mcool_files = []
        self.coolers = []
        self.resolution = resolution
   

    def extract_mcool_files(self):
        self.mcool_files = []
        self.mcool_names = []
        self.GSE = []
        for root, dir_, files in os.walk(self.directory):
            for f in files:
                if f.endswith('.mcool'):
                    self.mcool_files.append(root + '/' + f)
                    self.mcool_names.append(f.split('.')[0])
                    expri = root.split('_' +f.split('.')[0]+ '_tmp')[0].split('/')[-1]
                    clss = root.split(expri+'_'+f.split('.')[0]+'_tmp')[0].split('/')[-2]
                    self.GSE.append(clss)


    def read_mcool_files(self):
        for mcool_file in self.mcool_files:
            try:
                cooler_obj = cooler.Cooler(mcool_file+"::/resolutions/"+str(self.resolution))
                self.coolers.append(cooler_obj)
            except Exception as e:
                print(f"Error reading file {mcool_file}: {str(e)}")

    def count_values(self):
        ###此函数为了计算每个single cell HIC matrix 有值的个数
        counts = []
        chromCounts = []
        for cooler_obj in self.coolers:
            count = 0
            currentChromDict = {}
            for chrom in cooler_obj.chromnames:
                matrix = cooler_obj.matrix(balance=False).fetch(chrom)
                chromCount = np.count_nonzero(matrix)
                count += chromCount
                currentChromDict[chrom] = chromCount
            counts.append(count)
            chromCounts.append(currentChromDict)

        return counts, chromCounts

    def get_coolers(self):
        return self.coolers


    def count_distance_interaction(self, RandomSize=3):
        ###此函数为了计算每个single cell HIC matrix 远端有值的个数，先得到矩阵，再生成一个相同大小、离对角线越近越可能是0、越远越可能是1的矩阵，之后再相乘
        counts = []
        chromCounts = []
        chromLens = []
        for cooler_obj in self.coolers:
            count = 0
            currentChromDict = {}
            currentChromLen = {}
            for chrom in cooler_obj.chromnames:
                matrix = cooler_obj.matrix(balance=False).fetch(chrom)
                matrixLen = np.shape(matrix)[0]
                distanceCount=0
                for i in range(RandomSize):
                    randomMat = self.generate_random_matrix(matrixLen)
                    matrix = matrix * randomMat
                    distanceCount += np.count_nonzero(matrix)
                count += distanceCount / RandomSize
                currentChromDict[chrom] = distanceCount / RandomSize
                currentChromLen[chrom] = matrixLen
            counts.append(count)
            chromCounts.append(currentChromDict)
            chromLens.append(currentChromLen)
        return counts, chromCounts, chromLens


    def generate_random_matrix(self, size):
        # Generate a matrix of random values between 0 and 1
        random_matrix = np.random.rand(size, size)

        # Calculate the distances from the main diagonal for all elements
        distances = np.abs(np.arange(size) - np.arange(size)[:, np.newaxis])

        # Calculate the probabilities for all elements based on distances
        probabilities = 1.0 - (distances / (size - 1))

        probabilities = probabilities * probabilities * probabilities
        # Generate the final matrix based on probabilities
        matrix = (random_matrix < probabilities).astype(int)

        return matrix

    def calculate_Chorm_interactions(self):
        ###计算染色体间的interaction
        interactions = []
        for cooler_obj in self.coolers:
            chromnames = cooler_obj.chromnames
            num_chroms = len(chromnames)
            chromnamesInteraction = {}
            for i in range(num_chroms):
                for j in range(i, num_chroms):
                    inter = cooler_obj.matrix(balance=False).fetch(chromnames[i], chromnames[j])
                    interaction = np.count_nonzero(inter)
                    chromnamesInteraction[chromnames[i]+"-"+chromnames[j]] = interaction
            interactions.append(chromnamesInteraction)
        return interactions

    def calculateAjcMatEmbadding(self, similarityMethod = 'gaussian_distance', embeddingMethod = 'tSNE', dim = 2,cache=True):
        CellSimilarity = []
        CellFeature = []
        fileCount = len(self.coolers)
        if not cache or not os.path.exists('./cache/cache.npz'):
            # 不存在cache，那还是要做一遍，但是存下来
            for i in range(fileCount):
                ####获取该细胞的HIC矩阵
                TargetMatrix_1 = self.coolers[i].matrix(balance=False)[:]

                ###获取距离对角线0-10的数据并降成一维
                diagonal_data = [np.diag(TargetMatrix_1, k) for k in range(0, 10)]

                ###获取降维后的矩阵
                SampledTargetMatrix_1 = self.scale_matrix(TargetMatrix_1, 200)

                ###整合整体特征与染色体内部特征
                CellFeature.append(np.concatenate([np.concatenate([arr.flatten() for arr in diagonal_data]), SampledTargetMatrix_1.flatten()], axis=0))

                ####计算细胞间的similarity
                CurCellSimilarity = []
                for j in range(fileCount):
                    TargetMatrix_2 = self.coolers[j].matrix(balance=False)[:]
                    SampledTargetMatrix_2 = self.scale_matrix(TargetMatrix_2, 200)
                    if similarityMethod == 'gaussian_distance':
                        CurCellSimilarity.append(self.gaussian_distance(SampledTargetMatrix_1, SampledTargetMatrix_2, 1))
                    if similarityMethod == 'EuclideanDistance':
                        CurCellSimilarity.append(np.linalg.norm(SampledTargetMatrix_1 - SampledTargetMatrix_2))
                    if similarityMethod == 'CosineSimilarity':
                        CurCellSimilarity.append(1 - self.cosine_similarity(SampledTargetMatrix_1, SampledTargetMatrix_2))
                ###记录每个细胞与所有细胞间的similarity
                CellSimilarity.append(CurCellSimilarity)
        else:
            ###存在cache，那就直接读取
            print('load cache')
            cache = np.load('./cache/cache.npz')
            CellFeature = cache['CellFeature']
            CellSimilarity = cache['CellSimilarity']

        
        
        # print(CellFeature)
        ##对矩阵正态分布处理
        CellSimilarity = np.array(CellSimilarity)
        smoothed_CellSimilarity = 1 - norm.cdf(CellSimilarity, np.mean(CellSimilarity), np.std(CellSimilarity))
        smoothed_CellSimilarity = (smoothed_CellSimilarity - np.min(smoothed_CellSimilarity)) / (np.max(smoothed_CellSimilarity) - np.min(smoothed_CellSimilarity))

        ###计算每个细胞的embadding
        CellFeature = np.array(CellFeature)
        self.CellFeature = CellFeature
        self.cellSimilarity = smoothed_CellSimilarity
        if cache and not os.path.exists('./cache/cache.npz'):
            if not os.path.exists('cache'):
                os.makedirs('cache')
            print('save cache')
            cache = {}
            cache['CellFeature'] = CellFeature
            cache['CellSimilarity'] = smoothed_CellSimilarity
            cache['mcool_names'] = self.mcool_names
            cache['GSE'] = self.GSE
            np.savez('./cache/cache.npz', **cache)

        if embeddingMethod == 'PCA':
            CellsEmbedding = PCAEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'MDS':
            CellsEmbedding = MDSEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'tSNE':
            CellsEmbedding = tSNEEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'UMAP':
            CellsEmbedding = UMAPEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'LLE':
            CellsEmbedding = LLEEmbedding(CellFeature, dim=dim)

        if embeddingMethod == 'Isomap':
            CellsEmbedding = IsomapEmbedding(CellFeature, dim=dim)

        return smoothed_CellSimilarity, CellsEmbedding

    

    def gaussian_distance(self, matrix1, matrix2, sigma):
        # 计算高斯核矩阵
        kernel1 = np.exp(-np.sum((matrix1[:, np.newaxis] - matrix1) ** 2, axis=2) / (2 * sigma ** 2))
        kernel2 = np.exp(-np.sum((matrix2[:, np.newaxis] - matrix2) ** 2, axis=2) / (2 * sigma ** 2))

        # 计算高斯距离
        distance = np.sum((kernel1 - kernel2) ** 2)

        return distance


    def cosine_similarity(self, matrix_A, matrix_B):
        ####计算两个矩阵的余弦相似性
        vector_A = matrix_A.ravel()  # 使用ravel()代替flatten()，无需创建副本
        vector_B = matrix_B.ravel()

        dot_product = np.dot(vector_A, vector_B)
        norm_A = np.linalg.norm(vector_A)
        norm_B = np.linalg.norm(vector_B)

        similarity = dot_product / (norm_A * norm_B)
        return similarity

    def scale_matrix(self, matrix, target_size):
        n = matrix.shape[0]  # 原始矩阵的大小
        scale_factor = n / target_size  # 计算缩放比例

        # 构建目标矩阵的行和列索引
        row_indices = np.arange(target_size) * scale_factor
        col_indices = np.arange(target_size) * scale_factor

        # 使用索引进行插值计算
        row_floor = row_indices.astype(int)
        row_ceil = np.ceil(row_indices).astype(int)
        col_floor = col_indices.astype(int)
        col_ceil = np.ceil(col_indices).astype(int)

        # 计算四个邻近像素值
        Q11 = matrix[row_floor[:, None], col_floor]
        Q12 = matrix[row_floor[:, None], col_ceil]
        Q21 = matrix[row_ceil[:, None], col_floor]
        Q22 = matrix[row_ceil[:, None], col_ceil]

        # 双线性插值计算目标矩阵
        target_matrix = (Q11 * (row_ceil[:, None] - row_indices[:, None]) * (col_ceil - col_indices) +
                         Q21 * (row_indices[:, None] - row_floor[:, None]) * (col_ceil - col_indices) +
                         Q12 * (row_ceil[:, None] - row_indices[:, None]) * (col_indices - col_floor) +
                         Q22 * (row_indices[:, None] - row_floor[:, None]) * (col_indices - col_floor))

        return target_matrix



from sc3dg.analysis.api import MMCooler
import click

import cooler
import os

'''
    Traversing the directory, and mergeing the mcool files in the directory
'''


@click.command()
@click.option('--mcool', help='mcool directory path', required=True, default=None)
@click.option('--output', help='output path', required=True, default=None)
@click.option('--resolution', help='resolution', required=True, type=str)
@click.option('--num', help='top number', required=False, default=None, type=int)
@click.option('--sort', help='sort', required=False, default=True, type=bool)
@click.option('--n_jobs', help='n_jobs', required=False, default=10, type=int)

def  merge(mcool,
                           output,
                           resolution,
                           num,
                           sort=True,
                           n_jobs=10):
    '''
    Traversing the directory, and mergeing the mcool files in the directory,
    if num is larger than the number of the mcool files, then merge all the mcool files
    Parameters
    ----------
    mcool: str
        The directory path of the mcool files, traverse the directory and merge the mcool files
    output: str
        The output file path, which should be a file name
    resolution: list, like [10000, 40000, 100000, 500000, 1000000]
        The resolution of the mcool files

    num: int
        The number of the top level mcool files
    sort: bool
        Whether to sort the mcool files, default is True, merge Top level mcool files
        if false, merge the bottom level mcool files
    Returns
    -------
        None
    '''
    data = MMCooler(directory=mcool,
                    resolution=resolution,
                    merge=False,
                    n_jobs=n_jobs
                    )

    data.sorted_cell(num=num, reverse=sort)

    # save
    cooler.merge_coolers(input_uris=[x.uri for x in data.clr_obj[0]],
                         output_uri=output + '.cool',
                         mergebuf=1000000)

    cooler.zoomify_cooler(base_uris=output, outfile=output,
                          resolutions=resolution,
                          chunksize=1000000, nproc=16)
    os.remove(output + '.cool')











import gzip
import sys
import argparse
import re
import numpy as np
import subprocess
import time
import logging
import os


def pp(opt, fastq,log_out):

    
    print('********run sn_m3c*********')
    T = time.time()
    # 移动到工作目录
    os.chdir(opt['output'])

        # 创建文件夹tmp,并移动
    if not os.path.exists(opt['type'] + '_' + fastq + '_tmp'): # scHic_sample_tmp
        os.makedirs(opt['type'] + '_' + fastq + '_tmp')
    

    os.chdir(opt['type'] + '_' + fastq + '_tmp')
    if os.path.exists(opt['output'] + '/' + opt['type'] + '_' + fastq + '_tmp/Result/' + fastq +'.mcool'):
        log_out.info('The result of ' + fastq + ' has been generated')
        return
    else:
        for val in os.listdir(opt['output'] + '/' + opt['type'] + '_' + fastq + '_tmp'):
            print('rm -r ' + val)
            os.system('rm -r ' + val)
            print('clearing done')

    print('********run sn_m3c after clearing*********')

    
    fq_r1 = opt['fastq_dir'][opt['fastq_log'].index(fastq)]
    fq_r2 = fq_r1.replace('_1.fastq', '_2.fastq')


    # 创建logging文件
    logging.basicConfig(filename=fastq + '_logging.log', level=logging.DEBUG)
 


    for k in opt.keys():

     
        if (k == 'fastq_dir') or (k == 'fastq_log') or (k == 'run_sample'):
            continue
        else:
            logging.debug('# ' + k + ': ' + str(opt[k]))

    if not os.path.exists('Result'):
        os.makedirs('Result')
    if not os.path.exists('Result/SCPair'):
        os.makedirs('Result/SCPair')
    
    # Get paird fastq
    cmd = 'fastp '
    cmd += '--trim_front1 25 --trim_front2 25 --trim_tail1 3 --trim_tail2 3 '
    cmd += ' -i ' + fq_r1 + ' -I ' + fq_r2
    cmd += ' --out1 trimmed-pair1.fastq.gz --out2 trimmed-pair2.fastq.gz'
    cmd += ' --failed_out failed.fq.gz'
    cmd += ' --html trimmed.fastp.html '
    cmd += ' --json trimmed.fastp.json '
    print('1++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('fastp time: ' + str(time.time() - t))
    
    # Align reads to combo-reference using bwa/bowtie2
    opt['index'] = opt['index'].rsplit('/', 1)[0]

    t = time.time()
    if os.path.exists(opt['index'] + '/Bisulfite_Genome' ) :
        pass
    else:
        logging.info('create bismark index')
        cmd = 'bismark_genome_preparation --bowtie2 ' + opt['index']
        logging.info(cmd)
        os.system(cmd)
        logging.info('bismark_genome_preparation time: ' + str(time.time() - t))
   


    # ####第一次mapping
    cmd = 'bismark --bowtie2 ' + opt['index']
    cmd += ' --fastq trimmed-pair1.fastq.gz --pbat -un --parallel ' + str(opt['thread'])
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('bismark1 time: ' + str(time.time() - t))

    cmd = 'bismark --bowtie2 ' + opt['index']
    cmd += ' --fastq trimmed-pair2.fastq.gz -un --parallel ' + str(opt['thread'])
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('bismark2 time: ' + str(time.time() - t))

    # 对umapped的进行剪切继续map
    t = time.time()
    logging.info('split unmapped reads')
    split( 'trimmed-pair1.fastq.gz_unmapped_reads.fq.gz', 40 ,40)
    split( 'trimmed-pair2.fastq.gz_unmapped_reads.fq.gz', 40 ,40)
    logging.info('split time: ' + str(time.time() - t))


    # 对umapped进行mapped
    cmd = 'bismark --bowtie2 ' + opt['index']
    cmd += ' --fastq trimmed-pair1.fastq.gz_unmapped_reads.fq.gz_r1.fq --pbat --parallel ' + str(opt['thread'])
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('bismark3 time: ' + str(time.time() - t))

    cmd = 'bismark --bowtie2 ' + opt['index']
    cmd += ' --fastq trimmed-pair2.fastq.gz_unmapped_reads.fq.gz_r1.fq  --parallel ' + str(opt['thread'])
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('bismark4 time: ' + str(time.time() - t))


    # rm
    # cmd = 'rm *_unmapped_reads.fq.*'
    # logging.info(cmd)
    # t = time.time()
    # os.system(cmd)
    # logging.info('rm time: ' + str(time.time() - t))
    


    # merge
    cmd = 'samtools merge *.bam -o ' + fastq + '.merged.bam'
    logging.info(cmd)
    t = time.time()
    os.system(cmd)
    logging.info('samtools merge time: ' + str(time.time() - t))


    t = time.time()
    qc = opt['qc']
    cmd = 'samtools view -q %s -@ %s -b ' % (qc,str(opt['thread'])) + fastq + '.merged.bam -o ' + fastq + '.filtered.bam'
    logging.info(cmd)
    os.system(cmd)

    cmd = 'samtools sort -@ %s -n ' % str(opt['thread']) + fastq + '.filtered.bam -o ' + fastq + '.sorted.bam'
    logging.info(cmd)
    os.system(cmd)

    cmd = 'samtools view -@ %s -h ' % str(opt['thread']) + fastq + '.sorted.bam -o ' + fastq  + '.sorted.sam'
    logging.info(cmd)
    os.system(cmd)
    logging.info('samtools view time: ' + str(time.time() - t))

    logging.info('filtSam')
    t = time.time()
    filtSam(fastq + '.sorted.sam', fastq + '.sam')
    logging.info('filtSam time: ' + str(time.time() - t))


    # Generate pairs
    cmd = 'pairtools parse ' + fastq + '.sam ' + ' -c ' + opt['genomesize']
    cmd += ' --drop-sam --drop-seq --output-stats ' + fastq + '.stats '
    cmd += ' --assembly ' + opt['species']  + ' --no-flip '
    cmd += ' --add-columns ' + opt['add_columns']
    cmd += ' --walks-policy all '
    cmd += '-o ' + fastq + '.pairs.gz'
    cmd += ' --nproc-in ' + str(opt['thread']) +  ' --nproc-out ' + str(opt['thread'])
    # print('4++++++++++++\n',cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('pairtools parse time: ' + str(time.time() - t))

    cmd = 'pairtools restrict -f ' + opt['enzyme_bed'] + ' ' + fastq + '.pairs.gz -o ' + fastq + '.restrict.pairs.gz'
    # print('4++++++++++++\n',cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('pairtools restrict time: ' + str(time.time() - t))


    # #QC select
    cmd = 'pairtools select ' + '\"' + opt['select'] + '\"'
    cmd += ' ' + fastq + '.restrict.pairs.gz -o ' + fastq + '.filtered.pairs.gz'
    # print('5++++++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('pairtools select time: ' + str(time.time() - t))

    # Sort pairs
    cmd ='pairtools sort --nproc ' + str(opt['thread'])
    cmd += ' ' + fastq + '.filtered.pairs.gz -o ' + fastq + '.sorted.pairs.gz'
    print('6++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('pairtools sort time: ' + str(time.time() - t))

    #dedup
    cmd = 'pairtools dedup '
    cmd += '--max-mismatch ' + str(opt['max_mismatch'])
    cmd += ' --mark-dups --output ' + '>( pairtools split --output-pairs ' + fastq + \
        '.nodups.pairs.gz --output-sam ' + fastq + \
        '.nodups.bam ) --output-unmapped >( pairtools split --output-pairs  ' + fastq + \
        '.unmapped.pairs.gz --output-sam  ' + fastq + \
        '.unmapped.bam ) --output-dups >( pairtools split --output-pairs  ' + fastq + \
        '.dups.pairs.gz --output-sam ' + fastq + \
        '.dups.bam ) --output-stats  ' + fastq + \
        '.dedup.stats ' + fastq + \
        '.sorted.pairs.gz'
    # print('7+++++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    subprocess.run(cmd, shell=True, executable="/bin/bash")
    logging.info('pairtools dedup time: ' + str(time.time() - t))
    

    # #QC select
    cmd = 'pairtools select ' + '\"' + opt['select'] + '\"'
    cmd += ' ' + fastq + '.nodups.pairs.gz -o ' + fastq + '.nodups.UU.pairs.gz'
    # print('8++++++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('pairtools select time: ' + str(time.time() - t))


        # 
    cmd = 'gunzip ' + fastq + '.nodups.UU.pairs.gz'
    # print('9+++++++++++\n',cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('gunzip time: ' + str(time.time() - t))
    
    cmd = 'cooler cload pairs -c1 2 -p1 3 -c2 4 -p2 5 '
    cmd += opt['genomesize'] + ':' + str(opt['resolution']) + ' ' + fastq + '.nodups.UU.pairs ' + fastq + str(opt['resolution']) + '.cool'
    # print('10+++++++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('cooler cload time: ' + str(time.time() - t))
   
    cmd = 'gzip ' + fastq + '.nodups.UU.pairs'
    # print('11+++++++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('gzip time: ' + str(time.time() - t))

    # # KR correction
    # cmd = 'hicCorrectMatrix correct --matrix  ' + fastq + str(opt['resolution']) + '.cool '
    # cmd += ' --correctionMethod KR --outFileName '
    # cmd += ' ' + fastq + str(opt['resolution']) + '.KR.cool'
    # cmd += ' --filterThreshold -1.5 5.0 --chromosomes chr1 chr2 chr3 chr4 chr5 chr6 chr7 chr8 chr9 chr10 chr11 chr12 chr13 chr14 chr15 chr16 chr17 chr18 chr19 chrX '
    # # print('12+++++++++++\n', cmd)
    # t = time.time()
    # logging.info(cmd)
    # os.system(cmd)
    # logging.info('hicCorrectMatrix time: ' + str(time.time() - t))
   
    

    cmd = 'cooler zoomify '  + fastq + str(opt['resolution']) + '.cool -r %s -o ' % opt['zoomify_res'] + fastq + '.mcool'
    # print('13+++++++++++\n', cmd)
    t = time.time()
    logging.info(cmd)
    os.system(cmd)
    logging.info('cooler zoomify time: ' + str(time.time() - t))


    os.system('mv ' + fastq + '.mcool ./Result/')
    os.system('mv ' + fastq + str(opt['resolution']) + '.KR.cool ./Result/')
    os.system('mv ' + fastq + str(opt['resolution']) + '.cool ./Result/')
    os.system('mv ' + fastq + '.nodups.pairs.gz ./Result/')

    logging.info('total time: ' + str(time.time() - T))






def split(fn1, size1, size2):
    if 'gz' in fn1:
            dfh1=gzip.open(fn1,'r')

    if 'gz' not in fn1:
        dfh1=open(fn1,'r')
    if sys.version_info[0]==3:
        ID1=str(dfh1.readline().rstrip())[2:-1]
        line1=str(dfh1.readline().rstrip())[2:-1]
        plus1=str(dfh1.readline().rstrip())[2:-1]
        QS1=str(dfh1.readline().rstrip())[2:-1]
    if sys.version_info[0]==2:
        ID1=dfh1.readline().rstrip()
        line1=dfh1.readline().rstrip()
        plus1=dfh1.readline().rstrip()
        QS1=dfh1.readline().rstrip()
    rfhs=[]


    trim1=5
    trim2=5

    for i in range(1,4):
        rfhs.append(open(fn1+'_r'+str(i)+'.fq','w'))


    while line1:
        if len(line1[trim1:size1+trim1])>=30:
            rfhs[0].write(str(ID1)+'-1'+'\n'+str(line1[trim1:size1+trim1])+'\n'+str(plus1[0])+'\n'+str(QS1[trim1:size1+trim1])+'\n')
        if len(line1[trim1+size1:(-1*size2)-trim2])>=30:
            rfhs[0].write(str(ID1)+'-2'+'\n'+str(line1[trim1+size1:(-1*size2)-trim2])+'\n'+str(plus1[0])+'\n'+str(QS1[trim1+size1:(-1*size2)-trim2])+'\n')
        if len(line1[(-1*size2)-trim2:-1*trim2])>=30:
            rfhs[0].write(str(ID1)+'-3'+'\n'+str(line1[(-1*size2)-trim2:-1*trim2])+'\n'+str(plus1[0])+'\n'+str(QS1[(-1*size2)-trim2:-1*trim2])+'\n')
        if sys.version_info[0]==3:
            ID1=str(dfh1.readline().rstrip())[2:-1]
            line1=str(dfh1.readline().rstrip())[2:-1]
            plus1=str(dfh1.readline().rstrip())[2:-1]
            QS1=str(dfh1.readline().rstrip())[2:-1]
        if sys.version_info[0]==2:
            ID1=dfh1.readline().rstrip()
            line1=dfh1.readline().rstrip()
            plus1=dfh1.readline().rstrip()
            QS1=dfh1.readline().rstrip()  

    map(lambda x:x.close(),rfhs)


def filtSam(SamFile, CleanFile ):
  
    fp = open(CleanFile, "w")
    Read_list = []
    ReadID_list = []
    with open(SamFile, 'r') as file:
        lines = file.readlines()

    for i in range(len(lines)-1):
        if "@" in lines[i]:
            #print(lines[i])
            #print(i)
            fp.write(lines[i])

        else:
            Read_ID1 = lines[i].split("\t")[0].split("_")[0]
            Read_ID2 = lines[i+1].split("\t")[0].split("_")[0]
            components_1 = lines[i].split("\t")
            components_2 = lines[i+1].split("\t")

            if Read_ID1 == Read_ID2:
                components_1[0] = Read_ID1
                components_2[0] = Read_ID2
                lines1 = "\t".join(components_1)
                lines2 = "\t".join(components_2)

                fp.write(lines1+lines2)
                lines[i+1] = ''
    fp.close()




# Description: This script is used to make index for bowtie2 or bwa
import click
import time
from sc3dg.utils.download import make_index

@click.command()
@click.option('-g', '--genome', type=click.Choice(['hg38', 'hg19', 'mm10', 'mm9']), required=True, help='Choose from hg38, hg19, mm10, mm9')
@click.option('-a', '--aligner', type=click.Choice(['bowtie2', 'bwa']), required=True, help='Choose from bowtie2, bwa')
@click.option('-p', '--path', required=True, help='Path to save index')
def make_index(genome, aligner, path):
    make_index(path, genome, aligner)

if __name__ == '__main__':
    start_time = time.time()
    make_index()
    print('=============time spent:', time.time() - start_time, 'seconds===================')



from sc3dg.model.nuc_dynamics import calc_genome_structure
from time import time
import os
import click
def generate_3d(pair_file_path:str,
                output:str,
                num_models:int,
                iter_steps:int,
                iter_res:list,
                support_chrom:list):

    # Number of alternative conformations to generate from repeat calculations
    # with different random starting coordinates
    num_models = num_models

    # Parameters to setup restraints and starting coords
    general_calc_params = {'dist_power_law': -0.33,
                           'contact_dist_lower': 0.8, 'contact_dist_upper': 1.2,
                           'backbone_dist_lower': 0.1, 'backbone_dist_upper': 1.1,
                           'random_seed': int(time()), 'random_radius': 10.0}

    # Annealing & dyamics parameters: the same for all stages
    # (this is cautious, but not an absolute requirement)
    anneal_params = {'temp_start': 5000.0, 'temp_end': 10.0, 'temp_steps': 100,
                     'dynamics_steps': iter_steps, 'time_step': 0.001}
                     # ynamics_step 精细度
    # Hierarchical scale protocol: calculations will initially use 8 Mb particles
    # but deminish to 100 kb at the end. The whole annealing protocol (hot to cold)
    # will be run at each size, but subsequent stages will start from the previous
    # structure
    # particle_sizes = [8e6, 4e6, 2e6, 4e5, 2e5, 1e5]
    particle_sizes =  iter_res
    # Contacts must be clustered with another within this separation threshold
    # (at both ends) to be considered supported, i.e. not isolated
    # This removes noise contacts
    isolation_threshold = 2e6

    Target_Chrom = ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9', 'chr10', 'chr11', 'chr12',
                    'chr13', 'chr14', 'chr15', 'chr16',
                    'chr17', 'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrX']
    if os.path.exists(pair_file_path):
        import sys
        sys.exit("Output file already exists: %s" % pair_file_path)



    # Actually run the calculation with the specified parameters and input
    # The below function will automatically create the appropriate distance restraints
    calc_genome_structure(pair_file_path, output, general_calc_params, anneal_params,
                          particle_sizes, num_models, isolation_threshold, support_chrom=support_chrom, out_format='pdb')


@click.command('model')
@click.option('--pair', required=True, type=click.Path(exists=True), help='Path to the pair file.')
@click.option('--output', required=True, type=click.Path(), help='Path to the output pdb file.')
@click.option('--num_models', default=5, type=int, help='Number of models to generate.')
@click.option('--iter_steps', default=10, type=int, help='Number of iteration steps.')
@click.option('--iter_res', default="8e6,4e6,2e6,4e5,2e5,1e5", help='Comma-separated list of iteration resolutions.')
@click.option('--support_chrom', default=None, help='Comma-separated list of supported chromosomes.')
def model(pair, output, num_models, iter_steps, iter_res, support_chrom):
    """Model command to run the specified model with the config file."""
    iter_res_list = [float(res) for res in iter_res.split(',')]
    generate_3d(pair, output, num_models, iter_steps, iter_res_list)


import numpy as np
from scipy.sparse import csr_matrix, diags, eye
from scipy.sparse.linalg import norm
from scipy.ndimage import gaussian_filter
from functools import reduce
import cooler
import logging
import pandas as pd
import click

def calculate_sparsity(sparse_matrix):
    """Calculate the sparsity of a sparse matrix."""
    rows, cols = sparse_matrix.shape
    sparsity = sparse_matrix.nnz / (rows * cols)
    return sparsity

def perform_random_walk(matrix_P, restart_prob, tolerance):
    """Perform Random Walk with Restart (RWR) on a sparse matrix."""
    if restart_prob == 1:
        return matrix_P

    num_genes = matrix_P.shape[0]
    identity_matrix = eye(num_genes, dtype=np.float32)
    transition_matrix = matrix_P.copy()

    for i in range(100):
        new_transition_matrix = matrix_P.dot(transition_matrix * (1 - restart_prob) + restart_prob * identity_matrix)
        delta = norm(transition_matrix - new_transition_matrix)
        transition_matrix = new_transition_matrix.copy()
        if delta < tolerance:
            break

    return transition_matrix

def make_cooler_bins_chr(chr_name, length, resolution):
    total_bin = []
    for i in range(0, length, resolution):
        total_bin.append([chr_name, i, min(length, i+resolution)])
    df = pd.DataFrame(total_bin, columns=['chrom', 'start', 'end'])
    return df

# 其他杂染色体可以自行添加，字典里有就会读取没有就跳过
chr_list = ['chr' + str(i) for i in [k for k in range(1, 23)] + ['X', 'Y']]

def merge_segement(hic_dict,
                   save_path,
                   resolution, #分辨率要已知
                   chr_list = chr_list,
                   scale_size = 10000, # 概率矩阵上整体放大倍数
                  ):
    total_matrix_list = []
    total_bin_list = []
    stack = 0
    for chr_name in chr_list:
        if chr_name in hic_dict:
            bin_df = make_cooler_bins_chr(chr_name,
                                          length=hic_dict[chr_name].shape[0] * resolution,
                                          resolution=resolution)
            total_bin_list.append(bin_df)
            # 先算单独的坐标，再加stack
            chr_matrix = hic_dict[chr_name]
            rows, cols = np.nonzero(chr_matrix)
            #print(rows, cols)
            counts = np.array(hic_dict[chr_name][rows, cols]) * scale_size
            rows += stack; cols += stack
            matrix_sp = pd.DataFrame(np.column_stack((rows, cols, counts[0])))
            matrix_sp = matrix_sp[matrix_sp.iloc[:, 0] <= matrix_sp.iloc[:, 1]]
            total_matrix_list.append(matrix_sp)
            stack += bin_df.shape[0]

    bin_df_total = reduce(lambda x, y: pd.concat([x,y], axis = 0), total_bin_list)
    bin_df_total = bin_df_total.reset_index(drop=True)

    sp_df_total = reduce(lambda x, y: pd.concat([x,y], axis = 0), total_matrix_list)
    sp_df_total = sp_df_total.reset_index(drop=True)
    sp_df_total.columns = ['bin1_id', 'bin2_id', 'count']
    cooler.create_cooler(cool_uri=save_path,
                         bins=bin_df_total, pixels=sp_df_total)







@click.command()
@click.option('--mcool',  help='cool file path',required=True, default=None)
@click.option('--resolution', help='resolution',required=True, default=None, type=int)
@click.option('--output', help='output file path',required=True, default=None)
@click.option('--use_logscale', help='use logscale',required=False, default=False, type=bool)
@click.option('--padding', help='padding',required=False, default=1, type=int)
@click.option('--std_dev', help='std_dev',required=False, default=1, type=float)
@click.option('--restart_prob', help='restart_prob',required=False, default=0.5, type=float)
@click.option('--tolerance', help='tolerance',required=False, default=1e-6, type=float)

def impute(mcool, 
            resolution, 
            output, 
            use_logscale=False, 
            padding=1, 
            std_dev=1, 
            restart_prob=0.5,
        tolerance=1e-6):
    """
    Impute missing values in Hi-C data using Random Walk with Restart (RWR).

    Args:
        cool_file_url (str): URL or path to the .cool file.
        resolution (int): Resolution of the Hi-C data.
        output_file_path (str): Path to save the imputed matrices.
        use_logscale (bool): Whether to log-transform the data.
        padding (int): Padding for Gaussian filter.
        std_dev (float): Standard deviation for Gaussian filter.
        restart_prob (float): Restart probability for RWR.
        tolerance (float): Convergence tolerance for RWR.
        window_size_bp (int): Size of the window for local smoothing in base pairs.
        step_size_bp (int): Step size for sliding window in base pairs.
        output_distance_bp (int): Output distance for storing matrices in base pairs.
        min_cutoff (int): Minimum cutoff for values in the matrix.
    """
    if mcool is None:
        print("ERROR: Must provide a cool file URL or path.")
        return

    hic_data = cooler.Cooler(f"{mcool}::/resolutions/{resolution}")
    rwr_results = {}

    for chromosome in hic_data.chromnames:
        contact_matrix = hic_data.matrix(balance=False, sparse=True).fetch(chromosome)

        # Log transform
        if use_logscale:
            contact_matrix.data = np.log2(contact_matrix.data + 1)

        # Remove diagonal before convolution
        contact_matrix = contact_matrix - diags(contact_matrix.diagonal())

        # Apply Gaussian filter
        if padding > 0:
            contact_matrix = gaussian_filter(contact_matrix.astype(np.float32).toarray(), std_dev, order=0, mode='mirror', truncate=padding)
            contact_matrix = csr_matrix(contact_matrix)

        # Remove diagonal before RWR
        contact_matrix = contact_matrix - diags(contact_matrix.diagonal())

        # Prepare matrix for RWR
        matrix_B = contact_matrix + diags((contact_matrix.sum(axis=0).A.ravel() == 0).astype(int))
        degree_matrix = diags(1 / matrix_B.sum(axis=0).A.ravel())
        transition_probability_matrix = degree_matrix.dot(matrix_B).astype(np.float32)
        imputed_matrix = perform_random_walk(transition_probability_matrix, restart_prob, tolerance)

        # Normalize
        imputed_matrix += imputed_matrix.T
        degree = imputed_matrix.sum(axis=0).A.ravel()
        degree[degree == 0] = 1
        normalization_matrix = diags(1 / np.sqrt(degree))
        imputed_matrix = normalization_matrix.dot(imputed_matrix).dot(normalization_matrix)

        # Mask the lower triangle of the imputed matrix
        upper_triangle_indices = np.triu_indices(imputed_matrix.shape[0], 0)
        upper_triangle_mask = csr_matrix((np.ones(len(upper_triangle_indices[0])), (upper_triangle_indices[0], upper_triangle_indices[1])),
                                         imputed_matrix.shape, dtype=np.float32)
        imputed_matrix = imputed_matrix.tocsr().multiply(upper_triangle_mask)

        # Store RWR data
        rwr_results[chromosome] = imputed_matrix

    # Save the RWR results to cool output path
    merge_segement(rwr_results, save_path= output, resolution=resolution)

# Example usage:
# impute_hic_data(cool_file_url='/Users/jiangwenjie/D/ScHIC_pipeline/test.mcool', resolution=500000,output_file_path='/Users/jiangwenjie/D/ScHIC_pipeline/Code/test_RWR.cool')


import pandas as pd
import sys


def WriteCellPairs(ChromSize, SamFile, GenomeInfo, ResultDir):

    fp = open(ResultDir + "/" + str(SamFile).split("/")[-1].split(".")[0]+".pairs", "w")
    fp.write("## pairs format v1.0.0\n#shape: whole matrix\n#genome_assembly: " + GenomeInfo + "\n")
    for line in open(ChromSize):
        fp.write("#chromsize: " + line.split("\t")[0] + " " + line.split("\t")[1])
    

    Anchor_longreads = ''
    Anchor_longreads_pairsInfo = []
    flag=0
    for line in open(SamFile):

        ####Deal Sam file head
        if "@" == line[0]:
            fp.write("#samheader: " + line)

        else:
            if flag==0:
                fp.write("#columns: readID chrom1 pos1 chrom2 pos2 strand1 strand2 pair_type mapq1 mapq2\n")
                flag+=1
            ReadComponents = line.split("\t")

            ####判断正负链
            if bool(int(ReadComponents[1]) & 0x1000):
                ReadComponents[1] = '+'
            else:
                ReadComponents[1] = '-'

            current_longreads = ReadComponents[0].split(":")[0]

            if current_longreads == Anchor_longreads:
                Anchor_longreads_pairsInfo.append([ReadComponents[0], ReadComponents[2], ReadComponents[3], ReadComponents[1], ReadComponents[4]])

            else:
                for i in range(len(Anchor_longreads_pairsInfo)):
                    for j in range(0,i):
                        fp.write(Anchor_longreads_pairsInfo[i][0] + "_" + Anchor_longreads_pairsInfo[j][0] + "\t" +
                                 Anchor_longreads_pairsInfo[i][1] + "\t" + Anchor_longreads_pairsInfo[i][2] + "\t" +
                                 Anchor_longreads_pairsInfo[j][1] + "\t" + Anchor_longreads_pairsInfo[j][2] + "\t" +
                                 Anchor_longreads_pairsInfo[i][3] + "\t" + Anchor_longreads_pairsInfo[j][3] + "\tUU\t" +
                                 Anchor_longreads_pairsInfo[i][4] + "\t" + Anchor_longreads_pairsInfo[j][4] + "\n")

                Anchor_longreads = current_longreads
                Anchor_longreads_pairsInfo = []
                Anchor_longreads_pairsInfo.append([ReadComponents[0], ReadComponents[2], ReadComponents[3], ReadComponents[1], ReadComponents[4]])

    for i in range(len(Anchor_longreads_pairsInfo)):
        for j in range(0,i):
            fp.write(Anchor_longreads_pairsInfo[i][0] + "_" + Anchor_longreads_pairsInfo[j][0] + "\t" +
                     Anchor_longreads_pairsInfo[i][1] + "\t" + Anchor_longreads_pairsInfo[i][2] + "\t" +
                     Anchor_longreads_pairsInfo[j][1] + "\t" + Anchor_longreads_pairsInfo[j][2] + "\t" +
                     Anchor_longreads_pairsInfo[i][3] + "\t" + Anchor_longreads_pairsInfo[j][3] + "\tUU\t" +
                     Anchor_longreads_pairsInfo[i][4] + "\t" + Anchor_longreads_pairsInfo[j][4] + "\n")

    fp.close()












"""
---- COPYRIGHT ----------------------------------------------------------------
Copyright (C) 2020
Connor Horton (Harvard University)

---- LICENSE ------------------------------------------------------------------
This file is part of GiniQC.

GiniQC is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

GiniQC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along
with this software.  If not, see <http://www.gnu.org/licenses/>.
"""

SLOPE = -0.0579
INTERCEPT = 1.233

import numpy as np
import sys, cooler
from collections import defaultdict
import warnings
import cooltools
def get_max_aberration(cool):
    matrix_data = np.array(cool.matrix(as_pixels=True, balance=False, join=True)[:])
    matrix_bins = np.array(cool.bins()[:])

    chromsizes = {}
    for bin_ in matrix_bins:
        chrom, start, end = bin_[:3]
        chromsizes[chrom] = end+1

    chrom_counts = defaultdict(int)
    for line in matrix_data:
        chr1, start1, end1, chr2, start2, end2, count = line
        if chr1 == chr2:
            chrom_counts[chr1] += count
        else:
            chrom_counts[chr1] += count
            chrom_counts[chr2] += count

    for chrom in chrom_counts:
        chrom_counts[chrom] /= chromsizes[chrom]
    coverage = list(chrom_counts.values())
    median_coverage = np.median(coverage)

    max_aberration = 1
    for chrom in coverage:
        # if chrom != 'chrY':
        if chrom > median_coverage:
            max_aberration = max(max_aberration, chrom/median_coverage)
        else:
            max_aberration = max(max_aberration, median_coverage/chrom)

    return max_aberration

def gini(values):
    values.sort()
    numerator, denominator = 0, 0
    count = 1
    n = len(values)
    for i in values:
        numerator += (2*count-n-1)*i
        denominator += i
        count += 1
    denominator *= n
    return numerator/denominator

def adjust(gini_index, numreads):
    if gini_index > 1e-4:
        return np.arctan(gini_index - SLOPE*np.log(numreads) - INTERCEPT)/np.pi + 0.5
    else:
        return np.nan

def normalize_matrix(cool):
    # normalizes matrix by iterative correction
    # with warnings.catch_warnings():
    #     warnings.simplefilter("ignore")
    #     bias, stats = cooler.ice.iterative_correction(cool, store=True, max_iters=200)
    # generates an array containing number of contacts, normalized weights for upper triangle of matrix
    matrix_data = np.array(cool.matrix(as_pixels=True, join=True, balance=True)[:])
    # gives total number of reads
    total_sum = sum(matrix_data[:,-2])

    bins = {}
    count = 0

    nbins = cool.bins().shape[0]
    normalized = np.zeros((nbins, nbins), dtype=np.float64)

    last_chrom = 0
    first_chrom_bin = 0

    for bin_ in np.array(cool.bins()[:]):
        chrom, start, end, weight = bin_
        if chrom != last_chrom:
            # mark all of the cis bins as -1 in order to separate them out later
            for i in range(first_chrom_bin, count):
                for j in range(first_chrom_bin, count):
                    normalized[i,j] = -1
            # reset values for new chromosome
            first_chrom_bin = count
            last_chrom = chrom
        bins[(chrom, start, end)] = count
        count += 1
    # make sure cis bins are set at -1 for last chromosome
    for i in range(first_chrom_bin, count):
        for j in range(first_chrom_bin, count):
            normalized[i,j] = -1

    # select trans reads only
    cistotal = 0
    transtotal = 0
    for matrix_bin in matrix_data:
        chr1, start1, end1, chr2, start2, end2, count, weight = matrix_bin
        bin1 = bins[(chr1, start1, end1)]
        bin2 = bins[(chr2, start2, end2)]
        if chr1 != chr2:
            normalized[bin1, bin2] = weight
            transtotal += count
        else:
            cistotal += count

    #remove cis contacts, bottom diagonal before gini index calculation
    gini_array = []
    for i in range(nbins):
        for j in range(i+1, nbins):
            if normalized[i,j] >= 0:
                gini_array.append(normalized[i,j])
    return gini_array, total_sum, cistotal, transtotal

def main():
    matrix = cooler.Cooler(sys.argv[1] + '::resolutions/1000000') # .cool filehandle
    if not cooltools.lib.is_cooler_balanced(matrix):
        cooler.balance_cooler(matrix, store=True)
    # first normalize matrix
    normalized, total_reads, cis_reads, trans_reads = normalize_matrix(matrix)
    # find maximum chromosomal read count aberration
    max_aberration = get_max_aberration(matrix)
    # then calculate gini index
    if total_reads < 1:
        print("Input matrix %s is empty" % sys.argv[1])
    else:
        gini_idx = gini(normalized)
        print(type(gini_idx))
        adjusted_gini = adjust(gini_idx, total_reads)
        percent_cis = 100.0*cis_reads/total_reads
        # print Gini to summary text file
        out = open(sys.argv[2], 'w')
        out.write("Cell name:\t%s\n" % sys.argv[1].split(".cool")[0])
        out.write("Total reads:\t%d\n" % total_reads)
        out.write("Percentage of contacts in cis:\t%.2f\n" % percent_cis)
        out.write("Raw GiniQC:\t%.3f\n" % gini_idx)
        out.write("Adjusted GiniQC:\t%.3f\n" % adjusted_gini)
        out.write("Maximum chromosomal coverage aberration:\t%.2f\n" % max_aberration)
        out.close()

def calculate_giniqc(matrix):

    normalized, total_reads, cis_reads, trans_reads = normalize_matrix(matrix)
    if total_reads < 1:
        print("Input matrix %s is empty" % matrix)
    else:
        gini_idx = gini(normalized)
        adjusted_gini = adjust(gini_idx, total_reads)
        Raw_GiniQC = "%.3f" % gini_idx
        Adjusted_GiniQC = "%.3f" % adjusted_gini
        return Raw_GiniQC, Adjusted_GiniQC





import click
from sc3dg.utils import scSPRITE,sn_m3c,scNano,scMethyl
from sc3dg.utils import help
from sc3dg.utils import tools as tl
from multiprocessing import Process, Queue, Pool, Manager
from sc3dg.utils.assembly import *
Path = os.getcwd()
print(Path)
tech = {
        'scSPRITE':scSPRITE,
        
        'sn_m3c':sn_m3c,
        'scNano':scNano,
        'scMethyl': scMethyl
       }


    


@click.command()
@click.option('-o', '--output', help='Save path', required=True)
@click.option('-f', '--fastq', required=True, help='Fastq directory, run all if -s is not specified')
@click.option('--logging', help='Logging file path', default='./log.log')
@click.option('-t', '--type', required=True, type=click.Choice(['scHic', 'snHic', 'sciHic', 'scSPRITE', 'dipC', 'sn_m3c', 'HIRES', 'scNano', 'scMethyl', 'LiMAC', 'scCARE']), help='Type of Hi-C')
@click.option('-e', '--enzyme', help='Enzyme, e.g., mboi', required=True, multiple=True)
@click.option('-r', '--resolution', help='Resolution', default=10000, type=int)
@click.option('-i', '--index', help='BWA fa file/Bowtie fa file', required=True)
@click.option('-s', '--sample', help='Sample file, txt', default=None)
@click.option('--exist-barcode', is_flag=True, help='Barcode exists')
@click.option('--qc', type=int, default=0, help='Samtools view to QC')
@click.option('--add-columns', default='mapq', help=help.help['add_columns'])
@click.option('--thread', help='Thread count', type=int, default=20)
@click.option('--worker', help='Simultaneously run the worker of the pipeline', type=int, default=2)
@click.option('--select', help=help.help['select'], default="mapq1>=30 and mapq2>=30")
@click.option('--max-mismatch', help=help.help['max_mismatch'], type=int, default=3)
@click.option('--aligner', default='bwa', type=click.Choice(['bwa', 'bowtie2']), help='Aligner')
@click.option('--repeat-masker', help='Repeat masker (bed file) for scSPRITE', default='/cluster/home/Kangwen/common/mm10_rmsk.bed')
@click.option('--sprite-config', help='Sprite config', default=None)
@click.option('--scNano-barcode', help='scNano barcode for PCR and TN5, stored in a folder and named as TN5.txt and PCR index.txt respectively', default=None)
@click.option('--zoomify-res', help='zoomify',type=str, default='10000,40000,100000,500000,1000000')
def count(output, fastq, logging, type, enzyme, resolution, index, sample, exist_barcode, qc, add_columns, thread, worker, select, max_mismatch, aligner, repeat_masker, sprite_config, scNano_barcode, zoomify_res):
    """Run the Hi-C pipeline with the specified options."""
    opt = {
        'output': output,
        'fastq': fastq,
        'logging': logging,
        'type': type,
        'enzyme': enzyme,
        'resolution': resolution,
        'index': index,
        'sample': sample,
        'exist_barcode': exist_barcode,
        'qc': qc,
        'add_columns': add_columns,
        'thread': thread,
        'worker': worker,
        'select': select,
        'max_mismatch': max_mismatch,
        'aligner': aligner,
        'repeat_masker': repeat_masker,
        'sprite_config': sprite_config,
        'scNano_barcode': scNano_barcode,
        'zoomify_res': zoomify_res.split(',')
    }
    opt['running_path'] = Path

    # 先创建output文件夹，没有则创建，有则退出;然后移到该工作目录下面
    tl.make_workflow(opt['output'])
    os.chdir(opt['output'])

    # 然后判断是不是到底有多少个样本，如果是一个样本，那么就直接运行，如果是多个样本，那么就要用多进程来运行
    opt, fastq_log = tl.count_sample(opt)

 

    # 判断index是否存在
    tl.check_index(opt)
    tl.check_enzyme(opt)

    tl.get_enzyme_bed(opt)

    opt = tl.parser_fa_and_chrom_size(opt)

    # 查询是什么物种
    opt['species'] = tl.parse_species(opt)
    
    # print(opt)
    
    
    # gobal logging
    log_out = tl.log_(opt)
    os.chdir(opt['output'])

    if len(fastq_log) <= opt['worker']:
        print('*****num of worker is more than num of sample*****')
        p = Pool(opt['worker'])
        for fq in fastq_log:
            # 根据type来允许不同的流程
           run_exec(opt['type'],opt, fq,log_out)
        p.close()
        p.join()
    else:
        print('*****num of worker is less than num of sample*****')
        divide_list = []
        for i in range(0, len(fastq_log), opt['worker']):
            divide_list.append(fastq_log[i:i + opt['worker']])

        # logging batch 
        log_out.debug('divide_list:{}'.format(len(divide_list)))

        for i,val in enumerate(divide_list):
            p = Pool(len(val))
            for j,fq in enumerate(val):
                log_out.debug('%s || %s.....Dealing fq: %s \n' % (i*opt['worker']+ j,len(fastq_log),fq))
               
            
                p.apply_async(run_exec, args=(opt['type'],opt, fq,log_out))
                
            p.close()
            p.join()
            log_out.debug('Batch %s done || %s \n' % (i,len(divide_list)/opt['worker']))
    log_out.debug('All done || %s \n' % (len(fastq_log)))






import os
import sys

genome_url = {
    'mm10':'https://hgdownload.soe.ucsc.edu/goldenPath/mm10/bigZips/mm10.fa.gz',
    'mm9':'https://hgdownload.soe.ucsc.edu/goldenPath/mm9/bigZips/mm9.fa.gz',
    'mm39':'https://hgdownload.soe.ucsc.edu/goldenPath/mm39/bigZips/mm39.fa.gz',
    'hg38':'https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/hg38.fa.gz',
    'hg19':'https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/hg19.fa.gz'
}
chromosome_size_url = {
    'mm10':'https://hgdownload.soe.ucsc.edu/goldenPath/mm10/bigZips/mm10.chrom.sizes',
    'mm9':'https://hgdownload.soe.ucsc.edu/goldenPath/mm9/bigZips/mm9.chrom.sizes',
    'mm39':'https://hgdownload.soe.ucsc.edu/goldenPath/mm39/bigZips/mm39.chrom.sizes',
    'hg38':'https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/hg38.chrom.sizes',
    'hg19':'https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/hg19.chrom.sizes'
}


def make_index(path, version='hg38', aligner='bwa'):
    if not os.path.exists(path):
        os.makedirs(path)

    os.chdir(path)

    # 下载fa
    cmd = 'wget %s' % genome_url[version]
    os.system(cmd)

    cmd = 'wget %s' % chromosome_size_url[version]
    os.system(cmd)

    if aligner == 'bwa':
        cmd = 'bwa index %s.fa.gz' % version
        os.system(cmd)
    elif aligner == 'bowtie2':
        cmd = 'bowite2 index %s.fa.gz %s' % (version, version)
        os.system(cmd)
    else:
        sys.exit('aligner should be bwa/bowtie2')

    
def make_digest(path, enzyme='MboI',genome='hg38'):
    if not os.path.exists(path):
        sys.exit('path is not exist')
    print(os.path.exists(path + '/%s.fa' % genome))
    if not os.path.exists(path + '/%s.fa.gz' % genome) and not os.path.exists(path + '/%s.fa' % genome):
        sys.exit('you should run index first')
    if not os.path.exists(path + '/%s.fa.chrom.sizes' % genome):
        sys.exit('you should run index first')

    check_enzyme(enzyme)
    
    os.chdir(path)
    size = './%s.fa.chrom.sizes' % genome
    fa = './%s.fa.gz' % genome if os.path.exists('./%s.fa.gz' % genome) else './%s.fa' % genome
    
    cmd = 'cooler digest  -o digest_%s.bed %s %s %s' % (enzyme, size, fa, enzyme)
    print('digesting.....')
    os.system(cmd)
    

def check_enzyme(enzyme):
     import Bio.Restriction as biorst

     if enzyme not in biorst.AllEnzymes:
        lists = '['
        for val in list(biorst.AllEnzymes):
       
            lists += str(val )
            lists += ', '
        lists += ']'
        
        sys.exit('enzyme is not exist, you can use: ' + lists)







import click
from sc3dg.analysis.api import MMCooler



@click.command('gini')
@click.option('--mcool',  help='mcool path',required=True, default=None)
@click.option('--output',  help='output path',required=True, default=None)
@click.option('--resolution', help='resolution',required=True,  type=str)
def gini(mcool, output, resolution):
    '''
    Parameters
    ----------
    mcool:str
        mcool file path
    output:
        output file path
    resolution:
        resolution
    Returns
    -------
    '''
    ob1 = MMCooler(directory=mcool,
                   resolution=resolution.split(','),
                   describe = 'gini',
                   merge=False
                   )
    gini = ob1.get_gini()

    gini.to_csv(output, sep='\t')


import click
from sc3dg.commands.count import count
from sc3dg.commands.model import model

from sc3dg.commands.impute import impute
from sc3dg.commands.ssce import ssce
from sc3dg.commands.merge import merge
from sc3dg.commands.accum import accum
from sc3dg.commands.nDS import nDS
from sc3dg.commands.make_index import make_index


@click.group()
def cli():
    pass





cli.add_command(count)
cli.add_command(model)
cli.add_command(impute)
cli.add_command(ssce)
cli.add_command(merge)
cli.add_command(accum)
cli.add_command(nDS)
cli.add_command(make_index)


if __name__ == '__main__':
    cli()



import argparse
import re
import numpy as np
import subprocess
import time
import os
import logging
from .scaffold import *
import pandas as pd
from joblib import Parallel, delayed
def assembly(type_,opt, fastq,log_out):

    print('********run %s || %s*********' % (type_, fastq))
    # 移动到工作目录
    os.chdir(opt['output'])
    # print(opt)
    start = time.time()
 

    # 创建文件夹tmp,并移动
    if not os.path.exists(opt['type'] + '_' + fastq + '_tmp'): # scHic_sample_tmp
        os.makedirs(opt['type'] + '_' + fastq + '_tmp')
    os.chdir(opt['type'] + '_' + fastq + '_tmp')

   
    
    
    # 创建logging文件
    logging.basicConfig(filename=fastq + '_logging.log', level=logging.DEBUG)
    for k in opt.keys():

     
        if (k == 'fastq_dir') or (k == 'fastq_log') or (k == 'run_sample'):
            continue
        else:
            logging.debug('# ' + k + ': ' + str(opt[k]))
        
    if os.path.exists(opt['output'] + '/' + opt['type'] + '_' + fastq + '_tmp/' + fastq +'.bam'):
        log_out.info('The result of ' + fastq + ' has been generated')
        pass
        # print('==========generated+++++++')
        # return
    if os.path.exists(opt['output'] + '/' + opt['type'] + '_' + fastq + '_tmp/Result/SCpair'):
        for val in os.listdir(opt['output'] + '/' + opt['type'] + '_' + fastq + '_tmp/Result/SCpair'):
            if val.endswith('.mcool'):
                # print('==========generated+++++++')
                return

    
    
    


    fq_r1 = opt['fastq_dir'][opt['fastq_log'].index(fastq)]
    
    fq_r2 = fq_r1.replace('_1.fastq', '_2.fastq')
    
    if type_ == 'scHic' and opt['exist_barcode']:
        
        fq_r3 = fq_r1.replace('_1.fastq', '_3.fastq')
   
        fq_r4 = fq_r1.replace('_1.fastq', '_4.fastq')
   
        if not os.path.exists(fq_r3) or not os.path.exists(fq_r4):
            print(fastq , fq_r3, fq_r4)
            print('The barcode fastq file is not exist')
            log_out.error('The barcode fastq file is not exist')
            return
    
    threads = opt['thread']

    
    if not os.path.exists('Result'):
        os.makedirs('Result')
    


    # 默认inner和outer barcode和fastq文件放在一个目录下面
    if type_ == 'sciHic':
        opt['inner-barcode'] = os.path.dirname(opt['fastq_dir'][opt['fastq_log'].index(fastq)]) + '/' + fastq + '_inner_barcodes.txt'
        opt['outer-barcode'] = os.path.dirname(opt['fastq_dir'][opt['fastq_log'].index(fastq)]) + '/' + fastq + '_outer_barcodes.txt'



    # fastp,获得trimmed-pair1/2.fastq.gz,传入后续比对
    if opt['exist_barcode']:
        if type_ == 'snHic':
            t = fastp(fq_r1, fq_r2, 
                      'trimmed-pair1.fastq.gz', 'trimmed-pair2.fastq.gz',
                      threads, max=False)
            snHic_barcode('trimmed-pair1.fastq.gz', 'trimmed-pair2.fastq.gz')


        if type_ == 'scHic':
            print('making index..............')
            t  = GetFastqIndex(fastq, fq_r1, fq_r4, fq_r2, fq_r3)
            logging.info('make_index.py::: %s '%t)

            t = fastp('%s._1_barcode.fastq.gz'%fastq,
                    '%s._2_barcode.fastq.gz'%fastq, 
                    'trimmed-pair1.fastq.gz', 
                    'trimmed-pair2.fastq.gz',threads,  max=True)
            logging.info('fastp::: %s '%t)

        if type_ == 'sciHic':
            t = fastp(fq_r1,
                      fq_r2,
                      'trimmed-pair1_before.fastq.gz',
                        'trimmed-pair2_before.fastq.gz',
                        threads, max=False,
                        adapter=['AGATCGGAAGAGCGATCGG', 'AGATCGGAAGAGCGTCGTG']
                      )
            logging.info('fastp::: %s '%t)

            t = time.time()
            inline_splitter('trimmed-pair1_before.fastq.gz', 
                            'trimmed-pair2_before.fastq.gz',
                            opt['outer-barcode'], 
                            'trimmed-pair1_before2.fastq.gz', 
                            'trimmed-pair2_before2.fastq.gz',
                            fastq + '.splitting_stats.html')
            logging.info('inline_splitter::: %s '% (time.time()- t))

            t = time.time()
            analyze_scDHC_V2design(opt['inner-barcode'],
                            'trimmed-pair1_before2.fastq.gz', 
                            'trimmed-pair2_before2.fastq.gz',
                            'trimmed-pair1.fastq.gz', 
                            'trimmed-pair2.fastq.gz'
                            )
            logging.info('analyze_scDHC_V2design::: %s ' % (time.time()- t))

            t = time.time()
            awk()
            logging.info('awk::: %s ' % (time.time()- t))

    else:
        t = fastp(fq_r1, fq_r2, 'trimmed-pair1.fastq.gz', 'trimmed-pair2.fastq.gz',threads, max=False)
        logging.info('fastp::: %s '%t)
    

    # bwa,获得sam文件
    if opt['aligner'] == 'bwa':
        t = bwa(opt['index'], threads,fastq)
        logging.info('bwa::: %s '%t)
    else:
        t = bowtie2(opt['index'], threads,fastq)
        logging.info('bowtie2::: %s '%t)

    # # 比对的sam转bam
    t = sam_to_bam(fastq, threads)
    logging.info('sam_to_bam::: %s '%t)

    # # 质控bam文件
    t = samtools_qc(threads, opt['qc'], fastq)
    logging.info('samtools_qc::: %s '%t)

    # pairtools parse
    t = pairtools_parse(fastq, 
                        opt['genomesize'],
                        opt['species'],
                         opt['add_columns'],
                         threads
                        )
    logging.info('pairtools_parse::: %s '%t)


    # # pairtools restrict
    t = pairtools_restrict(fastq,opt['enzyme_bed'])
    logging.info('pairtools_restrict::: %s '%t)

    # # pairtools select
    t = pairtools_select(fastq, opt['select'])
    logging.info('pairtools_select::: %s '%t)

    # # pairtools sort 
    t = pairtools_sort(fastq)
    logging.info('pairtools_sort::: %s '%t)

    # pairtools dedup
    t = pairtools_dedup(fastq,str(opt['max_mismatch']))
    logging.info('pairtools_dedup::: %s '%t)

    # 整体的mcool
    t = cooler_cload_pairs(opt['genomesize'],
                            str(opt['resolution']),
                            fastq
                            ,prefix='dedup'
                            )
    logging.info('cooler_cload_pairs::: %s '%t)

    # t = KR_correctMatrix(fastq,str(opt['resolution']) )
    # logging.info('KR_correctMatrix::: %s '%t)

    t = cooelr_zoomify(fastq,opt['resolution'], opt['zoomify_res'])
    logging.info('cooelr_zoomify::: %s '%t)

    # print(34534534534)
    if opt['exist_barcode']:
        split_cells(fastq)
        t = time.time()
        pair_list = []
        for p in os.listdir('./Result/SCpair'):
            if p.endswith('pairs.gz'):
                pair_list.append(p)
    
        Parallel(n_jobs=6)(delayed(cload_correct_zoomify)('./Result/SCpair/',
                                      p,
                                      opt['genomesize'],
                                      str(opt['resolution']),
                                       opt['zoomify_res']
                            ) for p in pair_list)
        logging.info('sub_cload_correct_zoomify::: %s '% (time.time() - t))
 
    mv_Result()
    # pd.DataFrame(time_log, index=list(time_log.keys())).to_csv('time_log.csv', index=None, sep='\t')
    logging.info('total time: %s' % (time.time() - start))
    









    
    

    

        
        

    


from setuptools import setup, find_packages,Extension
from Cython.Build import cythonize
import numpy as np
numpy_includes = np.get_include()
with open('requirements.txt') as f:
    install_requires = f.read().splitlines()



setup(
    name="sc3dg",
    version="0.0.38",
    description="a toolkit for processing single cell Hi-C data",
    author="ABC",
    ext_modules=[
        Extension("sc3dg.model.dyn_util", ["sc3dg/model/dyn_util.c"], libraries=["m"], include_dirs=[numpy_includes]),
    ],
    packages=['sc3dg', 'sc3dg.utils', 'sc3dg.analysis',  'sc3dg.commands', 'sc3dg.model'],

    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'stark = sc3dg.cli:cli',
        ]}
)



from .analysis import MultiMcoolReader
from .plot import DrawCellInternerCount, DrawChromatainInternerCount, DrawDistanceCellInternerCount, DrawDistanceChromatainInternerCount
import numpy as np
import scanpy as sc
import matplotlib.pyplot as plt
import seaborn as sns
import os
import pandas as pd
import anndata as ad

def pp(opt):
    opt['output'] = opt['output'] + '/software_analysis'

    if not os.path.exists(opt['output']):
        os.makedirs(opt['output'])

    reader  = MultiMcoolReader(directory=opt['dir'], resolution=500000)
    print('extracting mcool files')
    reader.extract_mcool_files()
    print('reading mcool files')
    reader.read_mcool_files()
    print('getting cooler')
    coolers = reader.get_coolers()  
    print('calculating cell counts')
    countlist, ChromCountList = reader.count_values()
    print('calculating distance cell counts')
    distanceCountList, distanceChromCountList, chromLens = reader.count_distance_interaction()
    print('calculating chrom interaction')
    chromInteraction = reader.calculate_Chorm_interactions()
    print('calculating cell similarities')
    cellSimilarities, CellsEmbedding = reader.calculateAjcMatEmbadding()
    print('Drawing fig1-6')
    DrawCellInternerCount(CellCounts = countlist, ChromLenList = chromLens, save=opt['output']) 
    DrawChromatainInternerCount(ChromIntraction = chromInteraction, ChromLenList = chromLens, save=opt['output'])
    DrawDistanceCellInternerCount(CellCounts = distanceCountList, ChromLenList = chromLens, save=opt['output'])
    DrawDistanceChromatainInternerCount(ChromatainCounts=distanceChromCountList, save=opt['output'])

    print('Clustering')
    data = np.array(reader.CellFeature)
    adata = ad.AnnData(data)
    adata.obs['SRR'] = reader.mcool_names
    adata.obs['experiment'] = reader.GSE
    
    sc.pp.normalize_total(adata)
    sc.pp.log1p(adata)
    sc.tl.pca(adata, svd_solver='auto')
    sc.pp.neighbors(adata, n_neighbors=10, n_pcs=7)
    sc.tl.umap(adata)

    sc.tl.leiden(adata, resolution=opt['resolution'])
    for cr in ['experiment','leiden']:
        dot = adata.obsm['X_umap']
        if cr == 'leiden':
            label = list(np.unique(adata.obs[cr]).astype(int))
        else:
            label = list(np.unique(adata.obs[cr]))
        label = sorted(label)
        ax = plt.subplot(111)
        for cls in label:
            dot = adata.obsm['X_umap'][adata.obs[cr]==str(cls)]
            plt.scatter(dot[:,0],dot[:,1])
        plt.legend(label,loc='best')
        plt.xticks([])
        plt.yticks([])
        plt.xlabel('UMAP1')
        plt.ylabel('UMAP2')
        plt.savefig(opt['output'] + '/UMAP_{}_{}.png'.format(cr,opt['resolution']),dpi=300)
        plt.close()